#!/usr/bin/python
# -*- coding: utf-8 -*-

# provides functions for selecting a sample of training data
from __future__ import division

from collections import defaultdict
import itertools
from itertools import combinations, islice
from . import blocking, predicates, core, index
import numpy
import logging
import random

logger = logging.getLogger(__name__)

def findUncertainPairs(field_distances, data_model, bias=0.5):
    """
    Given a set of field distances and a data model return the
    indices of the record pairs in order of uncertainty. For example,
    the first indices corresponds to the record pair where we have the
    least certainty whether the pair are duplicates or distinct.
    """

    probability = core.scorePairs(field_distances, data_model)

    p_max = (1 - bias)
    logger.info(p_max)

    informativity = numpy.copy(probability)
    informativity[probability < p_max] /= p_max
    informativity[probability >= p_max] = (1 - probability[probability >= p_max])/(1-p_max)


    return numpy.argsort(-informativity)


class ActiveLearning(object) :
    """
    Ask the user to label the record pair we are most uncertain of. Train the
    data model, and update our uncertainty. Repeat until user tells us she is
    finished.
    """
    def __init__(self, candidates, data_model) :

        self.candidates = candidates
        self.field_distances = core.fieldDistances(candidates, data_model)
        self.seen_indices = set()

    def uncertainPairs(self, data_model, dupe_ratio) :
        uncertain_indices = findUncertainPairs(self.field_distances,
                                               data_model,
                                               dupe_ratio)

        for uncertain_index in uncertain_indices:
            if uncertain_index not in self.seen_indices:
                self.seen_indices.add(uncertain_index)
                break

        uncertain_pairs = [self.candidates[uncertain_index]]

        return uncertain_pairs


def semiSupervisedNonDuplicates(data_sample,
                                data_model,
                                nonduplicate_confidence_threshold=.9,
                                sample_size=2000):

    confidence = 1 - nonduplicate_confidence_threshold

    def distinctPairs() :
        data_slice = data_sample[0:sample_size]
        pair_distance = core.fieldDistances(data_slice, data_model)
        scores = core.scorePairs(pair_distance, data_model)

        sample_n = 0
        for score, pair in zip(scores, data_sample) :
            if score < confidence :
                yield pair
                sample_n += 1

        if sample_n < sample_size and len(data_sample) > sample_size :
            for pair in data_sample[sample_size:] :
                pair_distance = core.fieldDistances([pair], data_model)
                score = core.scorePairs(pair_distance, data_model)
                
                if score < confidence :
                    yield (pair)

    return islice(distinctPairs(), 0, sample_size)

def trainingData(training_pairs, record_ids) :

    record_pairs = set()
    tuple_pairs = set()
    for pair in training_pairs :
        record_pairs.add(tuple([(record_ids[record], record) 
                                for record in pair]))
        tuple_pairs.add(tuple([record_ids[record] 
                               for record in pair]))
    return record_pairs, tuple_pairs
    

def blockTraining(training_pairs,
                  predicate_set,
                  eta=.1,
                  epsilon=.1,
                  matching = "Dedupe"):
    '''
    Takes in a set of training pairs and predicates and tries to find
    a good set of blocking rules.
    '''

    if matching == "RecordLink" :
        Coverage = RecordLinkCoverage
    else :
        Coverage = DedupeCoverage

    # Setup
    try : # py 2
        record_ids = defaultdict(itertools.count().next) 
    except AttributeError : # py 3
        record_ids = defaultdict(itertools.count().__next__)

    dupe_pairs, training_dupes = trainingData(training_pairs['match'], 
                                              record_ids)

    distinct_pairs, training_distinct = trainingData(training_pairs['distinct'],
                                                     record_ids)
    
    coverage = Coverage(predicate_set,
                        dupe_pairs | distinct_pairs)

    predicate_set = coverage.overlap.keys()
    
    # Only consider predicates that cover at least one duplicate pair
    dupe_coverage = coverage.predicateCoverage(predicate_set,
                                               training_dupes)
    predicate_set = set(dupe_coverage.keys())

    # Throw away the predicates that cover too many distinct pairs
    coverage_threshold = eta * len(training_distinct)
    logger.info("coverage threshold: %s", coverage_threshold)

    distinct_coverage = coverage.predicateCoverage(predicate_set,
                                                   training_distinct)

    for pred, pairs in distinct_coverage.items() :
        if len(pairs) > coverage_threshold :
            predicate_set.remove(pred)

    chvatal_set, uncovered_dupes = findOptimumBlocking(training_dupes,
                                                       predicate_set,
                                                       distinct_coverage,
                                                       epsilon,
                                                       coverage)

    logger.debug("Uncovered Dupes")
    if uncovered_dupes :
        id_records = {v : k for k, v in record_ids.items()}
    for i, (rec_1, rec_2) in enumerate(uncovered_dupes) :
        logger.debug("uncovered pair %s" % (i,))
        logger.debug(id_records[rec_1])
        logger.debug(id_records[rec_2])
    

    final_predicate_set = removeSubsets(training_dupes,
                                        chvatal_set,
                                        coverage)

    logger.info('Final predicate set:')
    for predicate in final_predicate_set :
        logger.info(predicate)

    final_predicates = tuple(final_predicate_set)

    if final_predicate_set:
        return (final_predicates, 
                removeUnusedStopWords(coverage.stop_words,
                                      final_predicates))
    else:
        raise ValueError('No predicate found! We could not learn a single good predicate. Maybe give Dedupe more training data')


def findOptimumBlocking(uncovered_dupes,
                        predicate_set,
                        distinct_coverage,
                        epsilon,
                        coverage):

    # Greedily find the predicates that, at each step, covers the
    # most duplicates and covers the least distinct pairs, due to
    # Chvatal, 1979
    #
    # We would like to optimize the ratio of the probability of of a
    # predicate covering a duplicate pair versus the probability of
    # covering a distinct pair. If we have a uniform prior belief
    # about those probabilities, we can estimate these probabilities as
    #
    # (predicate_covered_dupe_pairs + 1) / (all_dupe_pairs + 2)
    #
    # (predicate_covered_distinct_pairs + 1) / (all_distinct_pairs + 2)
    #
    # When we are trying to find the best predicate among a set of
    # predicates, the denominators factor out and our coverage
    # estimator becomes
    #
    # (predicate_covered_dupe_pairs + 1)/ (predicate_covered_distinct_pairs + 1)

    dupe_coverage = coverage.predicateCoverage(predicate_set,
                                               uncovered_dupes)

    uncovered_dupes = set(uncovered_dupes)

    final_predicate_set = set()
    while len(uncovered_dupes) > epsilon:

        best_cover = 0
        best_predicate = None
        for predicate in dupe_coverage :
            dupes = len(dupe_coverage[predicate])
            distinct = len(distinct_coverage[predicate])
            cover = (dupes + 1)/(distinct + 1)
            if cover > best_cover:
                best_cover = cover
                best_predicate = predicate
                best_distinct = distinct
                best_dupes = dupes


        if not best_predicate:
            logger.warning('Ran out of predicates')
            break

        final_predicate_set.add(best_predicate)
        predicate_set.remove(best_predicate)
        
        uncovered_dupes = uncovered_dupes - dupe_coverage[best_predicate]
        dupe_coverage = coverage.predicateCoverage(predicate_set,
                                                   uncovered_dupes)


        logger.debug(best_predicate)
        logger.debug('cover: %(cover)f, found_dupes: %(found_dupes)d, '
                      'found_distinct: %(found_distinct)d, '
                      'uncovered dupes: %(uncovered)d',
                      {'cover' : best_cover,
                       'found_dupes' : best_dupes,
                       'found_distinct' : best_distinct,
                       'uncovered' : len(uncovered_dupes)
                       })

    return final_predicate_set, uncovered_dupes

def removeSubsets(uncovered_dupes, predicate_set, coverage) :
    dupe_coverage = coverage.predicateCoverage(predicate_set,
                                               uncovered_dupes)
    uncovered_dupes = set(uncovered_dupes)
    final_set = set()

    while uncovered_dupes :
        best_predicate = None
        max_cover = 0
        for predicate in dupe_coverage :
            cover = len(dupe_coverage[predicate])
            if cover > max_cover :
                max_cover = cover
                best_predicate = predicate

        if best_predicate is None :
            break

        final_set.add(best_predicate)
        predicate_set.remove(best_predicate)
        uncovered_dupes = uncovered_dupes - dupe_coverage[best_predicate]
        dupe_coverage = coverage.predicateCoverage(predicate_set,
                                                   uncovered_dupes)

    return final_set

class Coverage(object) :
    def __init__(self, predicate_set, pairs) :

        records = self._records_to_index(pairs)

        blocker = blocking.Blocker(predicate_set)

        for field, indices in blocker.index_fields.items() :
            record_fields = [record[field] 
                             for _, record 
                             in records
                             if record[field]]
            index_stop_words = stopWords(record_fields, indices) 
            blocker.stop_words[field].update(index_stop_words)
            blocker.index(sorted(set(record_fields)), field)

        self.stop_words = blocker.stop_words
        self.coveredBy(blocker.predicates, pairs)
        self.compoundPredicates()
        blocker.resetIndices()



    def coveredBy(self, predicates, pairs) :
        self.overlap = defaultdict(set)
        pairs = sorted(pairs)

        for predicate in predicates :
            rec_1 = None
            for pair in pairs :
                (record_1_id, record_1), (record_2_id, record_2) = pair
                if record_1_id != rec_1 :
                    blocks_1 = set(predicate(record_1))
                    rec_1 = record_1_id

                if blocks_1 :
                    blocks_2 = predicate(record_2)
                    field_preds = blocks_1 & set(blocks_2)
                    if field_preds :
                        rec_pair = record_1_id, record_2_id
                        self.overlap[predicate].add(rec_pair)



    def predicateCoverage(self,
                          predicate_set,
                          pairs) :

        coverage = defaultdict(set)
        pairs = set(pairs)

        for predicate in predicate_set :
            covered_pairs = pairs.intersection(self.overlap[predicate])
            if covered_pairs :
                coverage[predicate] = covered_pairs

        return coverage

    def compoundPredicates(self) :
        intersection = set.intersection

        compound_predicates = itertools.combinations(self.overlap, 2)

        for compound_predicate in compound_predicates :
            compound_predicate = predicates.CompoundPredicate(compound_predicate)
            self.overlap[compound_predicate] =\
                intersection(*[self.overlap[pred] 
                               for pred in compound_predicate])         


class DedupeCoverage(Coverage) :

    def _records_to_index(self, pairs) :
        return set(itertools.chain(*pairs))


class RecordLinkCoverage(Coverage) :

    def _records_to_index(self, pairs) :
        return {record_2 for _, record_2 in pairs}

def stopWords(data, indices) :
    index_stop_words = {}

    for index_type, predicates in indices.items() :
        processor = next(iter(predicates)).preprocess
        
        tf_index = index.CanopyIndex([])

        for i, doc in enumerate(data, 1) :
            tf_index.index_doc(i, processor(doc))

        doc_freq = [(len(tf_index.index._wordinfo[wid]), word) 
                    for word, wid in tf_index.lexicon.items()]

        doc_freq.sort(reverse=True)
        
        N = tf_index.index.documentCount()
        threshold = int(max(1000, N * 0.05))

        stop_words = set()
        
        for frequency, word in doc_freq :
            if frequency > threshold :
                stop_words.add(word)
            else :
                break

        index_stop_words[index_type] = stop_words

    return index_stop_words

def removeUnusedStopWords(stop_words, predicates) : # pragma : no cover
    new_dict = defaultdict(dict)
    for predicate in predicates :
        for pred in predicate :
            if hasattr(pred, 'index') :
                new_dict[pred.field][pred.type] = stop_words[pred.field][pred.type]

    return new_dict



from .base import FieldType
from dedupe import predicates
from haversine import haversine
import numpy

class LatLongType(FieldType) :
    type = "LatLong"

    _predicate_functions = [predicates.latLongGridPredicate]

    @staticmethod
    def comparator(field_1, field_2) :
        if field_1 == (0.0,0.0) or field_2 == (0.0,0.0) :
            return numpy.nan
        else :
            return haversine(field_1, field_2)




from .base import FieldType, DerivedType
from dedupe import predicates
from categorical import CategoricalComparator

class CategoricalType(FieldType) :
    type = "Categorical"
    _predicate_functions = [predicates.wholeFieldPredicate]

    def _categories(self, definition) :
        try :
            categories = definition["categories"]
        except KeyError :
            raise ValueError('No "categories" defined')
        
        return categories

    def __init__(self, definition) :

        super(CategoricalType, self ).__init__(definition)
        
        categories = self._categories(definition)

        self.comparator = CategoricalComparator(categories)
  
        self.higher_vars = []
        for higher_var in self.comparator.dummy_names :
            dummy_var = DerivedType({'name' : higher_var,
                                     'type' : 'Dummy',
                                     'has missing' : self.has_missing})
            self.higher_vars.append(dummy_var)

    def __len__(self) :
        return len(self.higher_vars)




from __future__ import division

from zope.index.text.lexicon import Lexicon
from zope.index.text.stopdict import get_stopdict
from zope.index.text.textindex import TextIndex
from zope.index.text.cosineindex import CosineIndex
from zope.index.text.setops import mass_weightedUnion

from BTrees.Length import Length
import re
import math
import numpy

class CanopyIndex(TextIndex) : # pragma : no cover
    def __init__(self, stop_words) : 
        lexicon = CanopyLexicon(stop_words)
        self.index = CosineIndex(lexicon)
        self.lexicon = lexicon

    def initSearch(self) :
        N = len(self.index._docweight)

        self._wids_dict = {}

        bucket = self.index.family.IF.Bucket
        for wid, docs in self.index._wordinfo.items() :
            if isinstance(docs, dict) :
                docs = bucket(docs)
            idf = numpy.log1p(N/len(docs))
            self.index._wordinfo[wid] = docs
            term = self.lexicon._words[wid]
            self._wids_dict[term] = (wid, idf)


    def apply(self, query_list, threshold, start=0, count=None):
        _wids_dict = self._wids_dict
        _wordinfo = self.index._wordinfo
        l_pow = float.__pow__

        L = []
        qw = 0

        for term in query_list :
            wid, weight = _wids_dict.get(term, (None, None))
            if wid is None :
                continue
            docs = _wordinfo[wid]
            L.append((docs, weight))
            qw += l_pow(weight, 2)

        results = mass_weightedUnion(L)

        qw = math.sqrt(qw)
        results = results.byValue(qw * threshold)

        return results


class CanopyLexicon(Lexicon) : # pragma : no cover
    def __init__(self, stop_words) : 
        super(CanopyLexicon, self).__init__()
        self._pipeline = [CustomStopWordRemover(stop_words)]

    def sourceToWordIds(self, last): 
        if last is None:
            last = [] 
        for element in self._pipeline:
            last = element.process(last)
        if not isinstance(self.wordCount, Length):
            self.wordCount = Length(self.wordCount())
        self.wordCount._p_deactivate()
        return list(map(self._getWordIdCreate, last))

class CustomStopWordRemover(object):
    def __init__(self, stop_words) :
        self.stop_words = set(get_stopdict().keys())
        self.stop_words.update(stop_words)

    def process(self, lst):
        return [w for w in lst if not w in self.stop_words]





from .base import FieldType
from dedupe import predicates

from affinegap import normalizedAffineGapDistance
from simplecosine.cosine import CosineTextSimilarity, CosineSetSimilarity

base_predicate_functions = (predicates.wholeFieldPredicate,
                            predicates.firstTokenPredicate,
                            predicates.commonIntegerPredicate,
                            predicates.nearIntegersPredicate,
                            predicates.firstIntegerPredicate,
                            predicates.sameThreeCharStartPredicate,
                            predicates.sameFiveCharStartPredicate,
                            predicates.sameSevenCharStartPredicate,
                            predicates.commonTwoTokens,
                            predicates.commonThreeTokens,
                            predicates.fingerprint,
                            predicates.oneGramFingerprint,
                            predicates.twoGramFingerprint,
                            predicates.sortedAcronym)

class ShortStringType(FieldType) :
    type = "ShortString"

    comparator = normalizedAffineGapDistance

    _predicate_functions = (base_predicate_functions 
                            + (predicates.commonFourGram,
                               predicates.commonSixGram,
                               predicates.tokenFieldPredicate,
                               predicates.suffixArray,
                               predicates.doubleMetaphone,
                               predicates.metaphoneToken))

    _index_predicates = (predicates.TfidfNGramCanopyPredicate, 
                         predicates.TfidfNGramSearchPredicate)
    _index_thresholds = (0.2, 0.4, 0.6, 0.8)

class StringType(ShortStringType) :
    type = "String"

    _index_predicates = (predicates.TfidfNGramCanopyPredicate, 
                         predicates.TfidfNGramSearchPredicate,
                         predicates.TfidfTextCanopyPredicate, 
                         predicates.TfidfTextSearchPredicate)


class TextType(FieldType) :
    type = "Text"

    _predicate_functions = base_predicate_functions 

    _index_predicates = (predicates.TfidfTextCanopyPredicate, 
                         predicates.TfidfTextSearchPredicate)
    _index_thresholds = (0.2, 0.4, 0.6, 0.8)

    def __init__(self, definition) :
        super(TextType, self).__init__(definition)


        if 'corpus' not in definition :
            definition['corpus'] = []

        self.comparator = CosineTextSimilarity(definition['corpus'])



#!/usr/bin/python
# -*- coding: utf-8 -*-

from collections import defaultdict
import collections
import itertools
import logging
import time

logger = logging.getLogger(__name__)

    

class Blocker:
    '''Takes in a record and returns all blocks that record belongs to'''
    def __init__(self, 
                 predicates, 
                 stop_words = None) :

        if stop_words is None :
            stop_words = defaultdict(lambda : defaultdict(set))

        self.predicates = predicates

        self.stop_words = stop_words

        self.index_fields = defaultdict(lambda:defaultdict(set))

        for full_predicate in predicates :
            for predicate in full_predicate :
                if hasattr(predicate, 'index') :
                    self.index_fields[predicate.field][predicate.type].add(predicate)

    #@profile
    def __call__(self, records):

        start_time = time.clock()
        predicates = [(':' + str(i), predicate)
                      for i, predicate
                      in enumerate(self.predicates)]

        

        for i, record in enumerate(records) :
            record_id, instance = record

            for pred_id, predicate in predicates :
                block_keys = predicate(instance)
                for block_key in block_keys :
                    yield block_key + pred_id, record_id
            
            if i and i % 10000 == 0 :
                logger.info('%(iteration)d, %(elapsed)f2 seconds',
                             {'iteration' :i,
                              'elapsed' :time.clock() - start_time})



    def resetIndices(self) :
        # clear canopies to reduce memory usage
        for index_type in self.index_fields.values() :
            for predicate in list(index_type.values())[0] :
                predicate.index = None
                if hasattr(predicate, 'canopy') :
                    predicate.canopy = {}

    def index(self, data, field): 
        '''Creates TF/IDF index of a given set of data'''
        indices = extractIndices(self.index_fields[field],
                                 self.stop_words[field])

        for doc in data :
            for _, index, preprocess in indices :
                index.index(preprocess(doc))

        for index_type, index, _ in indices :

            index.initSearch()

            for predicate in self.index_fields[field][index_type] :
                logger.info("Canopy: %s", str(predicate))
                predicate.index = index

    def unindex(self, data, field) :
        '''Remove index of a given set of data'''
        indices = extractIndices(self.index_fields[field])

        for doc in data :
            for _, index, preprocess in indices :
                index.unindex(preprocess(doc))

        for index_type, index, _ in indices :

            index._index.initSearch()

            for predicate in self.index_fields[field][index_type] :
                logger.info("Canopy: %s", str(predicate))
                predicate.index = index


def extractIndices(index_fields, stop_words=None) :
    
    indices = []
    for index_type, predicates in index_fields.items() :
        predicate = next(iter(predicates))
        index = predicate.index
        preprocess = predicate.preprocess
        if predicate.index is None :
            index = predicate.initIndex(stop_words[index_type])
        indices.append((index_type, index, preprocess))

    return indices




#!/usr/bin/python
# -*- coding: utf-8 -*-
import logging
from .index import CanopyIndex
import collections
import itertools

logger = logging.getLogger(__name__)

class TfIdfIndex(object) :
    def __init__(self, stop_words=[]) :
        self._index = CanopyIndex(stop_words)
 
        try : # py 2
            self._doc_to_id = collections.defaultdict(itertools.count(1).next)
        except AttributeError : # py 3
            self._doc_to_id = collections.defaultdict(itertools.count(1).__next__)

        
        self._parseTerms = self._index.lexicon.parseTerms

    def index(self, doc) :
        i = self._doc_to_id[doc]
        self._index.index_doc(i, doc)

    def unindex(self, doc) :
        i = self._doc_to_id.pop(doc)
        self._index.unindex_doc(i)
        self.initSearch()

    def initSearch(self) :
        self._index.initSearch()

    def search(self, doc, threshold=0) :
        query_list = self._parseTerms(doc)
 
        if query_list :
            results = [center for score, center 
                       in self._index.apply(query_list, threshold)]
        else :
            results = []

        return results




from dedupe.api import StaticDedupe, Dedupe
from dedupe.api import StaticRecordLink, RecordLink
from dedupe.api import StaticGazetteer, Gazetteer
from dedupe.core import randomPairs, randomPairsMatch, frozendict
from dedupe.convenience import consoleLabel, trainingDataDedupe, trainingDataLink, canonicalize



import simplejson as json
from simplejson.scanner import py_make_scanner
import dedupe.core

def _from_json(json_object):                                   
    if '__class__' in json_object:                            
        if json_object['__class__'] == 'frozenset':
            return frozenset(json_object['__value__'])
        if json_object['__class__'] == 'tuple':
            return tuple(json_object['__value__'])
    return json_object

class dedupe_encoder(json.JSONEncoder):

    def default(self, python_object):
        if isinstance(python_object, frozenset):                                
            python_object = {'__class__': 'frozenset',
                    '__value__': list(python_object)}
        if isinstance(python_object, dedupe.core.frozendict) :
            python_object = dict(python_object)
        
        return python_object

    def encode(self, python_object):
        if isinstance(python_object, tuple) :
            python_object = {'__class__': 'tuple',
                    '__value__': list(python_object)}
        
        return python_object

class dedupe_decoder(json.JSONDecoder):

    def __init__(self, **kwargs):
        json._toggle_speedups(False) # in simplejson, without this
                                     # some strings can be bytestrings
                                     # instead of unicode
                                     # https://code.google.com/p/simplejson/issues/detail?id=40
        json.JSONDecoder.__init__(self, object_hook=_from_json, **kwargs)
        # Use the custom JSONArray
        self.parse_array = self.JSONArray
        # Use the python implemenation of the scanner
        self.scan_once = py_make_scanner(self) 

    def JSONArray(self, s_and_end, scan_once, **kwargs):
        values, end = json.decoder.JSONArray(s_and_end, scan_once, **kwargs)
        return tuple(values), end




import pkgutil

from collections import OrderedDict

import dedupe.variables
import dedupe.variables.base as base
from dedupe.variables.base import MissingDataType
from dedupe.variables.interaction import InteractionType

for _, module, _  in pkgutil.iter_modules(dedupe.variables.__path__, 
                                          'dedupe.variables.') :
    __import__(module)


FIELD_CLASSES = dict(base.allSubclasses(base.FieldType))

class DataModel(dict) :

    def __init__(self, fields):

        self['bias'] = 0

        primary_fields, data_model = typifyFields(fields)
        self.derived_start = len(data_model)

        data_model += interactions(fields, primary_fields)
        data_model += missing(data_model)

        self['fields'] = data_model
        self.n_fields = len(self['fields'])
        self.primary_fields = primary_fields

    # Changing this from a property to just a normal attribute causes
    # pickling problems, because we are removing static methods from
    # their class context. This could be fixed by defining comparators
    # outside of classes in fieldclasses
    @property 
    def field_comparators(self) :
        start = 0
        stop = 0
        comparators = []
        for field in self.primary_fields :
            stop = start + len(field) 
            comparators.append((field.field, field.comparator, start, stop))
            start = stop

        return comparators

    @property 
    def missing_field_indices(self) : 
        return [i for i, definition 
                in enumerate(self['fields'])
                if definition.has_missing]

    @property
    def interactions(self) :
        indices = []

        fields = self['fields']
        field_names = [field.name for field in fields]

        for definition in fields :
            if hasattr(definition, 'interaction_fields') :
                interaction_indices = []
                for interaction_field in definition.interaction_fields :
                    interaction_indices.append(field_names.index(interaction_field))
                indices.append(interaction_indices)
                
        return indices

def typifyFields(fields) :
    primary_fields = []
    data_model = []

    for definition in fields :
        try :
            field_type = definition['type']
        except TypeError :
            raise TypeError("Incorrect field specification: field "
                            "specifications are dictionaries that must "
                            "include a type definition, ex. "
                            "{'field' : 'Phone', type: 'String'}")
        except KeyError :
            raise KeyError("Missing field type: fields "
                           "specifications are dictionaries that must "
                           "include a type definition, ex. "
                           "{'field' : 'Phone', type: 'String'}")
            
        if field_type == 'Interaction' :
            continue

        try :
            field_class = FIELD_CLASSES[field_type]
        except KeyError :
            raise KeyError("Field type %s not valid. Valid types include %s"
                           % (definition['type'], ', '.join(FIELD_CLASSES)))

        field_object = field_class(definition)
        primary_fields.append(field_object)
        
        if hasattr(field_object, 'higher_vars') :
            data_model.extend(field_object.higher_vars)
        else :
            data_model.append(field_object)

    return primary_fields, data_model

def missing(data_model) :
    missing_variables = []
    for definition in data_model[:] :
        if definition.has_missing :
            missing_variables.append(MissingDataType(definition.name))

    return missing_variables

def interactions(definitions, primary_fields) :
    field_d = {field.name : field for field in primary_fields}
    interaction_class = InteractionType

    interactions = []

    for definition in definitions :
        if definition['type'] == 'Interaction' :
            field = interaction_class(definition)
            field.expandInteractions(field_d)
            interactions.extend(field.higher_vars)

    return interactions






from .base import DerivedType
from categorical import CategoricalComparator
from .categorical_type import CategoricalType
from dedupe import predicates

class ExistsType(CategoricalType) :
    type = "Exists"
    _predicate_functions = []

    def __init__(self, definition) :

        super(CategoricalType, self ).__init__(definition)
        
        self.cat_comparator = CategoricalComparator([0,1])
  
        self.higher_vars = []
        for higher_var in self.cat_comparator.dummy_names :
            dummy_var = DerivedType({'name' : higher_var,
                                     'type' : 'Dummy',
                                     'has missing' : self.has_missing})
            self.higher_vars.append(dummy_var)

    def comparator(self, field_1, field_2) :
        if field_1 and field_2 :
            return self.cat_comparator(1, 1)
        elif field_1 or field_2 :
            return self.cat_comparator(0, 1)
        else :
            return self.cat_comparator(0, 0)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import print_function
from builtins import input

import collections
import itertools
import random
from dedupe.core import randomPairs
from canonicalize.centroid import getCanonicalRep

def consoleLabel(deduper): # pragma : no cover
    '''
    Command line interface for presenting and labeling training pairs
    by the user
    
    Argument :
    A deduper object
    '''

    finished = False

    while not finished :
        uncertain_pairs = deduper.uncertainPairs()

        labels = {'distinct' : [], 'match' : []}

        for record_pair in uncertain_pairs:
            label = ''
            labeled = False

            for pair in record_pair:
                for field in set(field[0] for field 
                                 in deduper.data_model.field_comparators) :
                    line = "%s : %s" % (field, pair[field])
                    print(line)
                print() 

            print('Do these records refer to the same thing?')
            valid_response = False
            while not valid_response:
                label = input('(y)es / (n)o / (u)nsure / (f)inished\n')
                if label in ['y', 'n', 'u', 'f']:
                    valid_response = True

            if label == 'y' :
                labels['match'].append(record_pair)
                labeled = True
            elif label == 'n' :
                labels['distinct'].append(record_pair)
                labeled = True
            elif label == 'f':
                print('Finished labeling')
                finished = True
            elif label != 'u':
                print('Nonvalid response')
                raise

        if labeled :
            deduper.markPairs(labels)
        

def trainingDataLink(data_1, data_2, common_key, training_size=50000) : # pragma : nocover
    '''
    Construct training data for consumption by the ActiveLearning 
    markPairs method from already linked datasets.
    
    Arguments : 
    data_1        -- Dictionary of records from first dataset, where the keys
                     are record_ids and the values are dictionaries with the 
                     keys being field names

    data_2        -- Dictionary of records from second dataset, same form as 
                     data_1
    
    common_key    -- The name of the record field that uniquely identifies 
                     a match
    
    training_size -- the rough limit of the number of training examples, 
                     defaults to 50000
    
    Warning:
    
    Every match must be identified by the sharing of a common key. 
    This function assumes that if two records do not share a common key 
    then they are distinct records. 
    '''
    
    
    identified_records = collections.defaultdict(lambda: [[],[]])
    matched_pairs = set()
    distinct_pairs = set()

    for record_id, record in data_1.items() :
        identified_records[record[common_key]][0].append(record_id)

    for record_id, record in data_2.items() :
        identified_records[record[common_key]][1].append(record_id)

    for keys_1, keys_2 in identified_records.values() :
        if keys_1 and keys_2 :
            matched_pairs.update(itertools.product(keys_1, keys_2))

    distinct_pairs = set(itertools.product(data_1.keys(), data_2.keys()))
    distinct_pairs -= matched_pairs
    distinct_pairs = random.sample(distinct_pairs, training_size)

    matched_records = [(data_1[key_1], data_2[key_2])
                       for key_1, key_2 in matched_pairs]

    distinct_records = [(data_1[key_1], data_2[key_2])
                        for key_1, key_2 in distinct_pairs]

    training_pairs = {'match' : matched_records, 
                      'distinct' : distinct_records} 

    return training_pairs        
        
        
def trainingDataDedupe(data, common_key, training_size=50000) : # pragma : nocover
    '''
    Construct training data for consumption by the ActiveLearning 
    markPairs method from an already deduplicated dataset.
    
    Arguments : 
    data          -- Dictionary of records, where the keys are record_ids and 
                     the values are dictionaries with the keys being 
                     field names

    common_key    -- The name of the record field that uniquely identifies 
                     a match
    
    training_size -- the rough limit of the number of training examples, 
                     defaults to 50000
    
    Warning:
    
    Every match must be identified by the sharing of a common key. 
    This function assumes that if two records do not share a common key 
    then they are distinct records. 
    '''

    
    identified_records = collections.defaultdict(list)
    matched_pairs = set()
    distinct_pairs = set()
    unique_record_ids = set()
    
    # a list of record_ids associated with each common_key
    for record_id, record in data.items() :
        unique_record_ids.add(record_id)
        identified_records[record[common_key]].append(record_id)

    # all combinations of matched_pairs from each common_key group
    for record_ids in identified_records.values() :
        if len(record_ids) > 1 :
            matched_pairs.update(itertools.combinations(sorted(record_ids), 2))

    # calculate indices using dedupe.core.randomPairs to avoid 
    # the memory cost of enumerating all possible pairs
    unique_record_ids = list(unique_record_ids)
    pair_indices = randomPairs(len(unique_record_ids), training_size)
    distinct_pairs = set()
    for i, j in pair_indices:
        distinct_pairs.add((unique_record_ids[i],
                            unique_record_ids[j]))

    distinct_pairs -= matched_pairs

    matched_records = [(data[key_1], data[key_2])
                       for key_1, key_2 in matched_pairs]

    distinct_records = [(data[key_1], data[key_2])
                        for key_1, key_2 in distinct_pairs]

    training_pairs = {'match' : matched_records, 
                      'distinct' : distinct_records} 

    return training_pairs


def canonicalize(record_cluster): # pragma : nocover
    """
    Constructs a canonical representation of a duplicate cluster by
    finding canonical values for each field

    Arguments:
    record_cluster     --A list of records within a duplicate cluster, where 
                         the records are dictionaries with field 
                         names as keys and field values as values

    """
    return getCanonicalRep(record_cluster)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from future.utils import viewitems

import itertools

import warnings
import numpy
import fastcluster
import hcluster

def connected_components(edgelist, max_components) :

    root = {}
    component = {}
    indices = {}

    if len(edgelist['pairs']) == 0:
        raise StopIteration()

    it = numpy.nditer(edgelist['pairs'], ['external_loop'])

    for i, (a,b) in enumerate(it) :
        root_a = root.get(a)
        root_b = root.get(b)

        if root_a is None and root_b is None :
            component[a] = {a, b}
            indices[a] = [i]
            root[a] = root[b] = a
        elif root_a is None or root_b is None :
            if root_a is None :
                a, b = b, a
                root_a, root_b = root_b, root_a
            component[root_a].add(b)
            indices[root_a].append(i)
            root[b] = root_a
        elif root_a != root_b :
            component_a = component[root_a]
            component_b = component[root_b]
            if len(component_a) < len(component_b) :
                root_a, root_b = root_b, root_a
                component_a, component_b = component_b, component_a

            component_a |= component_b
            indices[root_a].extend(indices[root_b])
            indices[root_a].append(i)

            for node in component_b :
                root[node] = root_a

            del component[root_b]
            del indices[root_b]
        else : 
            indices[root_a].append(i)

    for root in component :
        n_components = len(component[root])
        sub_graph = edgelist[indices[root]]
        
        if n_components > max_components :
            threshold = numpy.min(sub_graph['score'])
            threshold *= 1.1 
            warnings.warn('A component contained %s elements. '
                          'Components larger than %s are '
                          're-filtered. The threshold for this '
                          'filtering is %s' % (n_components, 
                                               max_components,
                                               threshold)) 
            filtered_sub_graph = sub_graph[sub_graph['score'] > threshold]	
            for sub_graph in connected_components(filtered_sub_graph, 
                                                  max_components) :
               yield sub_graph
        else :
            yield sub_graph
     

def condensedDistance(dupes):
    '''
    Convert the pairwise list of distances in dupes to "condensed
    distance matrix" required by the hierarchical clustering
    algorithms. Also return a dictionary that maps the distance matrix
    to the record_ids.
   
    The formula for an index of the condensed matrix is

    index = {N choose 2}-{N-row choose 2} + (col-row-1)
          = N*(N-1)/2 - (N-row)*(N-row-1)/2 + col - row - 1
            ^^^^^^^^^   ^^^^^^^^^^^^^^^^^^^   
          matrix_length       row_step
    
    where (row,col) is index of an uncondensed square N X N distance matrix.
    
    See http://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.squareform.html
    '''

    candidate_set = numpy.unique(dupes['pairs'])
    candidate_set = numpy.sort(candidate_set)

    i_to_id = dict(enumerate(candidate_set))

    ids = candidate_set.searchsorted(dupes['pairs'])
    row = ids[:, 0]
    col = ids[:, 1]

    N = len(numpy.union1d(row, col))
    matrix_length = N * (N - 1) / 2

    row_step = (N - row) * (N - row - 1) / 2
    index = matrix_length - row_step + col - row - 1

    condensed_distances = numpy.ones(matrix_length, 'f4')
    condensed_distances[index.astype(int)] = 1 - dupes['score']


    return i_to_id, condensed_distances, N


def cluster(dupes, threshold=.5, max_components=30000):
    '''
    Takes in a list of duplicate pairs and clusters them in to a
    list records that all refer to the same entity based on a given
    threshold

    Keyword arguments:
    threshold -- number betweent 0 and 1 (default is .5). lowering the 
                 number will increase precision, raising it will increase
                 recall
    '''
    threshold = 1 - threshold

    dupe_sub_graphs = connected_components(dupes, max_components)

    clustering = {}
    cluster_id = 0
    for sub_graph in dupe_sub_graphs:
        if len(sub_graph) > 1:

            i_to_id, condensed_distances, N = condensedDistance(sub_graph)

            linkage = fastcluster.linkage(condensed_distances,
                                          method='centroid', 
                                          preserve_input=False)

            partition = hcluster.fcluster(linkage, 
                                          threshold,
                                          criterion='distance')

            clusters = {}

            for (i, sub_cluster_id) in enumerate(partition):
                clusters.setdefault(cluster_id + sub_cluster_id, []).append(i)

            for cluster_id, items in viewitems(clusters) :
                if len(items) > 1 :
                    scores = confidences(items, condensed_distances, N)
                    clustering[cluster_id] =\
                        (tuple(i_to_id[item] for item in items), tuple(scores))

            cluster_id += max(partition) + 1
        else:
            ids, score = sub_graph[0]
            clustering[cluster_id] = (tuple(ids), tuple([score]*2))
            cluster_id += 1
            

    return clustering.values()

def confidences(items, condensed_distances, d) :
    scores = dict.fromkeys(items, 0)
    for i, j in itertools.combinations(items, 2) :
        index = d*(d-1)/2 - (d-i)*(d-i-1)/2 + j - i - 1
        dist = condensed_distances[index]
        scores[i] += dist
        scores[j] += dist
    scores = numpy.array([v for (k, v) in sorted(scores.items())])
    scores /= len(items) - 1
    scores = 1 - scores
    return scores

def greedyMatching(dupes, threshold=0.5):
    dupes = numpy.array(dupes)
    covered_vertex_A = set()
    covered_vertex_B = set()
    clusters = []

    sorted_dupes = sorted(dupes, key=lambda score: score[1], reverse=True)
    dupes_list = (dupe for dupe in sorted_dupes if dupe[1] >= threshold)

    for vertices, score in dupes_list:
        a, b = vertices
        if a not in covered_vertex_A and b not in covered_vertex_B:
            clusters.append((vertices, score))
            covered_vertex_A.add(a)
            covered_vertex_B.add(b)

    return clusters

def gazetteMatching(dupes, threshold=0.5, n_matches=1):
    dupes = numpy.array(dupes) 
    clusters = []

    sorted_dupes = sorted(dupes, key=lambda pair: (pair[0][0], -pair[1]))
    dupes_list = [dupe for dupe in sorted_dupes if dupe[1] >= threshold]

    if dupes_list :
        group = dupes_list[0][0][0]
        matches = []
        i = 0

        for pair, score in dupes_list:
            a, b = pair
            if a == group :
                if i < n_matches :
                    matches.append((pair, score))
                    i += 1
            else :
                clusters.append(tuple(matches))
                matches = [(pair, score)]
                i = 1
                group = a

        clusters.append(tuple(matches))

    return clusters



from .base import FieldType
from dedupe import predicates
import numpy

class ExactType(FieldType) :
    _predicate_functions = [predicates.wholeFieldPredicate]
    type = "Exact"

    @staticmethod
    def comparator(field_1, field_2) :
        if field_1 and field_2 :
            if field_1 == field_2 :
                return 1
            else :
                return 0
        else :
            return numpy.nan



import numpy
import weakref
import threading
import warnings
import platform
import sys

from future.utils import viewitems

MULTIPROCESSING = True
# Deal with Mac OS X issuse
config_info = str([value for key, value in
                   viewitems(numpy.__config__.__dict__)
                   if key.endswith("_info")]).lower()

if "accelerate" in config_info or "veclib" in config_info :
    warnings.warn("NumPy linked against 'Accelerate.framework'. "
                  "Multiprocessing will be disabled."
                  " http://mail.scipy.org/pipermail/numpy-discussion/2012-August/063589.html")
    MULTIPROCESSING = False
elif platform.system() == 'Windows' :
    warnings.warn("Dedupe does not currently support multiprocessing on Windows")
    MULTIPROCESSING = False

if MULTIPROCESSING :        
    from multiprocessing import Process, Pool, Queue
    if sys.version < '3':
        from multiprocessing.queues import SimpleQueue
    else :
        from multiprocessing import SimpleQueue
else :
    if not hasattr(threading.current_thread(), "_children"): 
        threading.current_thread()._children = weakref.WeakKeyDictionary()
    from multiprocessing.dummy import Process, Pool, Queue
    SimpleQueue = Queue





#!/usr/bin/python
# -*- coding: utf-8 -*-
from builtins import str

import re
import math
import itertools

from metaphone import doublemetaphone
from dedupe.cpredicates import ngrams, initials
import dedupe.tfidf as tfidf

words = re.compile(r"[\w']+").findall
integers = re.compile(r"\d+").findall
start_word = re.compile(r"^([\w']+)").match
start_integer = re.compile(r"^(\d+)").match

class Predicate(object) :
    def __iter__(self) :
        yield self
        
    def __repr__(self) :
        return "%s: %s" % (self.type, self.__name__)

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other) :
        return repr(self) == repr(other)

class SimplePredicate(Predicate) :
    type = "SimplePredicate"

    def __init__(self, func, field) :
        self.func = func
        self.__name__ = "(%s, %s)" % (func.__name__, field)
        self.field = field

    def __call__(self, record) :
        column = record[self.field]
        if column :
            return self.func(column)
        else :
            return ()

class ExistsPredicate(Predicate) :
    type = "ExistsPredicate"

    def __init__(self, field) :
        self.__name__ = "(Exists, %s)" % (field,)
        self.field = field


    @staticmethod
    def func(column) :
        if column :
            return ('1',)
        else :
            return ('0',)


    def __call__(self, record) :
        column = record[self.field]
        return self.func(column)


class IndexPredicate(Predicate) :
    def __init__(self, threshold, field):
        self.__name__ = '(%s, %s)' % (threshold, field)
        self.field = field
        self.threshold = threshold
        self.index = None

    def __getstate__(self):

        result = self.__dict__.copy()

        return {'__name__': result['__name__'],
                'field' : result['field'],
                'threshold' : result['threshold']}

    def __setstate__(self, d) :
        self.__dict__ = d
        self.index = None

class TfidfIndexPredicate(IndexPredicate) :
    def initIndex(self, stop_words) :
        return tfidf.TfIdfIndex(stop_words)
    

class TfidfSearchPredicate(TfidfIndexPredicate):
    def __call__(self, record) :
        column = record[self.field]
        if column :
            try :
                centers = self.index.search(self.preprocess(column), 
                                            self.threshold)
            except AttributeError :
                raise AttributeError("Attempting to block with an index "
                                     "predicate without indexing records")

            l_str = str
            return [l_str(center) for center in centers]

        else :
            return ()

class TfidfCanopyPredicate(TfidfIndexPredicate):
    def __init__(self, *args, **kwargs) :
        super(TfidfCanopyPredicate, self).__init__(*args, **kwargs)
        self.canopy = {}

    def __setstate__(self, *args, **kwargs) :
        super(TfidfCanopyPredicate, self).__setstate__(*args, **kwargs)
        self.canopy ={}

    def __call__(self, record) :
        block_key = None
        column = record[self.field]

        if column :

            doc = self.preprocess(column)

            try :
                doc_id = self.index._doc_to_id[doc]
            except AttributeError :
                raise AttributeError("Attempting to block with an index "
                                     "predicate without indexing records")

            if doc_id in self.canopy :
                block_key = self.canopy[doc_id]
            else :
                canopy_members = self.index.search(doc,
                                                   self.threshold)
                for member in canopy_members :
                    if member not in self.canopy :
                        self.canopy[member] = doc_id

                if canopy_members :
                    block_key = doc_id

        if block_key is None :
            return []
        else :
            return [str(block_key)]


class TfidfTextPredicate(object) :
    rx = re.compile(r"(?u)\w+[\w*?]*")

    def preprocess(self, doc) :
        return tuple(self.rx.findall(doc))

class TfidfSetPredicate(object) :
    def preprocess(self, doc) :
        return doc

class TfidfNGramPredicate(object) :
    def preprocess(self, doc) :
        return tuple(ngrams(doc.replace(' ', ''), 2))

class TfidfTextSearchPredicate(TfidfTextPredicate, 
                               TfidfSearchPredicate) :
    type = "TfidfTextSearchPredicate"

class TfidfSetSearchPredicate(TfidfSetPredicate, 
                              TfidfSearchPredicate) :
    type = "TfidfSetSearchPredicate"

class TfidfNGramSearchPredicate(TfidfNGramPredicate, 
                                TfidfSearchPredicate) :
    type = "TfidfNGramSearchPredicate"

class TfidfTextCanopyPredicate(TfidfTextPredicate, 
                               TfidfCanopyPredicate) :
    type = "TfidfTextCanopyPredicate"

class TfidfSetCanopyPredicate(TfidfSetPredicate, 
                              TfidfCanopyPredicate) :
    type = "TfidfSetCanopyPredicate"

class TfidfNGramCanopyPredicate(TfidfNGramPredicate, 
                                TfidfCanopyPredicate) :
    type = "TfidfNGramCanopyPredicate"



class CompoundPredicate(Predicate) :
    type = "CompoundPredicate"

    def __init__(self, predicates) :
        self.predicates = predicates
        self.__name__ = u'(%s)' % u', '.join([str(pred)
                                              for pred in 
                                              predicates])

    def __iter__(self) :
        for pred in self.predicates :
            yield pred

    def __call__(self, record) :
        predicate_keys = [predicate(record)
                          for predicate in self.predicates]
        return [u':'.join(block_key)
                for block_key
                in itertools.product(*predicate_keys)]
 

def wholeFieldPredicate(field):
    """return the whole field"""
    return (str(field), )

def tokenFieldPredicate(field):
    """returns the tokens"""
    return set(words(field))

def firstTokenPredicate(field) :
    first_token = start_word(field)
    if first_token :
        return first_token.groups()
    else :
        return ()

def commonIntegerPredicate(field):
    """return any integers"""
    return set(integers(field))

def nearIntegersPredicate(field):
    """return any integers N, N+1, and N-1"""
    ints = integers(field)
    near_ints = set(ints)
    for char in ints :
        num = int(char)
        near_ints.add(str(num-1))
        near_ints.add(str(num+1))
        
    return near_ints

def firstIntegerPredicate(field) :
    first_token = start_integer(field)
    if first_token :
        return first_token.groups()
    else :
        return ()


def ngramsTokens(field, n) :
    grams = set()
    n_tokens = len(field)
    for i in range(n_tokens):
        for j in range(i+n, min(n_tokens, i+n)+1):
            grams.add(' '.join(str(tok) for tok in field[i:j]))
    return grams


def commonTwoTokens(field) :
    return ngramsTokens(field.split(), 2)

def commonThreeTokens(field) :
    return ngramsTokens(field.split(), 3)

def fingerprint(field) :
    return (u''.join(sorted(field.split())).strip(),)

def oneGramFingerprint(field) :
    return (u''.join(sorted(set(ngrams(field.replace(' ', ''), 1)))).strip(),)

def twoGramFingerprint(field) :
    if len(field) > 1 :
        return (u''.join(sorted(gram.strip() for gram 
                                in set(ngrams(field.replace(' ', ''), 2)))),)
    else :
        return ()
    
def commonFourGram(field):
    """return 4-grams"""
    return set(ngrams(field.replace(' ', ''), 4))

def commonSixGram(field):
    """return 6-grams"""
    return set(ngrams(field.replace(' ', ''), 6))

def sameThreeCharStartPredicate(field):
    """return first three characters"""
    return initials(field.replace(' ', ''), 3)

def sameFiveCharStartPredicate(field):
    """return first five characters"""
    return initials(field.replace(' ', ''), 5)

def sameSevenCharStartPredicate(field):
    """return first seven characters"""
    return initials(field.replace(' ',''), 7)

def suffixArray(field) :
    field = field.replace(' ', '')
    n = len(field) - 4
    if n > 0 :
        for i in range(0, n) :
            yield field[i:]

def sortedAcronym(field) :
    return (''.join(sorted(each[0] for each in field.split())),)

def doubleMetaphone(field) :
    return [metaphone for metaphone in doublemetaphone(field) if metaphone]

def metaphoneToken(field) :
    return {metaphone_token for metaphone_token 
            in itertools.chain(*(doublemetaphone(token) 
                                 for token in set(field.split())))
            if metaphone_token}

def existsPredicate(field) :
    try :
        if any(field) :
            return (u'1',)
        else :
            return (u'0',)
    except TypeError :
        if field :
            return (u'1',)
        else :
            return (u'0',)

def wholeSetPredicate(field_set):
    return (str(field_set),)

def commonSetElementPredicate(field_set):
    """return set as individual elements"""
    return tuple([str(each) for each in field_set])

def commonTwoElementsPredicate(field) :
    l = sorted(field)
    return ngramsTokens(l, 2)

def commonThreeElementsPredicate(field) :
    l = sorted(field)
    return ngramsTokens(l, 3)

def lastSetElementPredicate(field_set) :
    return (str(max(field_set)), )

def firstSetElementPredicate(field_set) :
    return (str(min(field_set)), )

def magnitudeOfCardinality(field_set) :
    return orderOfMagnitude(len(field_set))

def latLongGridPredicate(field, digits=1):
    """
    Given a lat / long pair, return the grid coordinates at the
    nearest base value.  e.g., (42.3, -5.4) returns a grid at 0.1
    degree resolution of 0.1 degrees of latitude ~ 7km, so this is
    effectively a 14km lat grid.  This is imprecise for longitude,
    since 1 degree of longitude is 0km at the poles, and up to 111km
    at the equator. But it should be reasonably precise given some
    prior logical block (e.g., country).
    """
    if any(field) :
        return (str([round(dim, digits) for dim in field]),)
    else :
        return ()

def orderOfMagnitude(field) :
    if field > 0 :
        return (str(int(round(math.log10(field)))), )
    else :
        return ()

def roundTo1(field) : # thanks http://stackoverflow.com/questions/3410976/how-to-round-a-number-to-significant-figures-in-python
    abs_num = abs(field)
    order = int(math.floor(math.log10(abs_num)))
    rounded = round(abs_num, -order)
    return (str(int(math.copysign(rounded, field))),)
        



from dedupe import predicates

class Variable(object) :
    def __len__(self) :
        return 1

    def __repr__(self) :
        return self.name

    def __hash__(self) :
        return hash(self.name)

    def __eq__(self, other) :
        return self.name == other.name

    def __init__(self, definition) :

        self.weight = 0

        if definition.get('has missing', False) :
            self.has_missing = True
            try :
                self.predicates += [predicates.ExistsPredicate(definition['field'])]
            except AttributeError :
                pass
        else :
            self.has_missing = False

class DerivedType(Variable) :
    type = "Derived"

    def __init__(self, definition) :
        self.name = "(%s: %s)" % (str(definition['name']), 
                                  str(definition['type']))
        super(DerivedType, self).__init__(definition)


class MissingDataType(Variable) :
    type = "MissingData"

    def __init__(self, name) :
        
        self.name = "(%s: Not Missing)" % name
        self.weight = 0

        self.has_missing = False

class FieldType(Variable) :
    _index_thresholds = []
    _index_predicates = []

    def __init__(self, definition) :
        self.field = definition['field']

        if 'variable name' in definition :
            self.name = definition['variable name'] 
        else :
            self.name = "(%s: %s)" % (self.field, self.type)

        self.predicates = [predicates.SimplePredicate(pred, self.field) 
                           for pred in self._predicate_functions]

        self.predicates += indexPredicates(self._index_predicates,
                                           self._index_thresholds,
                                           self.field)


        super(FieldType, self).__init__(definition)

class CustomType(FieldType) :
    type = "Custom"
    _predicate_functions = []

    def __init__(self, definition) :
        super(CustomType, self).__init__(definition)

        try :
            self.comparator = definition["comparator"]
        except KeyError :
            raise KeyError("For 'Custom' field types you must define "
                           "a 'comparator' function in the field "
                           "definition. ")

        if 'variable name' in definition :
            self.name = definition['variable name'] 
        else :
            self.name = "(%s: %s, %s)" % (self.field, 
                                          self.type, 
                                          self.comparator.__name__)


def allSubclasses(cls) :
    field_classes = {}
    for q in cls.__subclasses__() :
        yield q.type, q
        for p in allSubclasses(q) :
            yield p

def indexPredicates(predicates, thresholds, field) :
    index_predicates = []
    for predicate in predicates :
        for threshold in thresholds :
            index_predicates.append(predicate(threshold, field))

    return index_predicates

    



from .base import FieldType
from dedupe import predicates
from simplecosine.cosine import CosineSetSimilarity

class SetType(FieldType) :
    type = "Set"

    _predicate_functions = (predicates.wholeSetPredicate,
                            predicates.commonSetElementPredicate,
                            predicates.lastSetElementPredicate,
                            predicates.commonTwoElementsPredicate,
                            predicates.commonThreeElementsPredicate,
                            predicates.magnitudeOfCardinality,
                            predicates.firstSetElementPredicate)
    
    _index_predicates = (predicates.TfidfSetSearchPredicate,
                         predicates.TfidfSetCanopyPredicate)
    _index_thresholds = (0.2, 0.4, 0.6, 0.8)

    def __init__(self, definition) :
        super(SetType, self).__init__(definition)

        if 'corpus' not in definition :
            definition['corpus'] = [] 

        self.comparator = CosineSetSimilarity(definition['corpus'])





from __future__ import division
from builtins import range, zip
from future.utils import viewitems

from collections import deque
import random
import functools
import itertools
import warnings
from collections import defaultdict

def blockedSample(sampler, sample_size, predicates, *args) :
    
    blocked_sample = set()
    remaining_sample = sample_size - len(blocked_sample)
    previous_sample_size = 0

    while remaining_sample and predicates :
        random.shuffle(predicates)

        new_sample = sampler(remaining_sample, 
                             predicates,
                             *args)

        filtered_sample = (subsample for subsample 
                           in new_sample if subsample)

        blocked_sample.update(itertools.chain.from_iterable(filtered_sample))

        growth = len(blocked_sample) - previous_sample_size
        growth_rate = growth/remaining_sample

        remaining_sample = sample_size - len(blocked_sample)
        previous_sample_size = len(blocked_sample)

        if growth_rate < 0.001 :
            warnings.warn("%s blocked samples were requested, "
                          "but only able to sample %s"
                          % (sample_size, len(blocked_sample)))
            break

            
        predicates = [pred for pred, pred_sample 
                      in zip(predicates, new_sample)
                      if pred_sample or pred_sample is None]
        
    return blocked_sample

def dedupeSamplePredicates(sample_size, predicates, items) :
    n_items = len(items)

    for subsample_size, predicate in subsample(sample_size, predicates) : 

        if not subsample_size :
            yield None
            continue

        items.rotate(random.randrange(n_items))
        items.reverse()

        yield dedupeSamplePredicate(subsample_size,
                                    predicate,
                                    items)

def dedupeSamplePredicate(subsample_size, predicate, items) :

    sample = []
    block_dict = {}
    predicate_function = predicate.func
    field = predicate.field

    for pivot, (index, record) in enumerate(items) :
        column = record[field]
        if not column :
            continue

        if pivot == 10000:
            if len(block_dict) + len(sample) < 10 :
                return sample

        
        block_keys = predicate_function(record[field])
        
        for block_key in block_keys:
            if block_key not in block_dict :
                block_dict[block_key] = index
            else :
                pair = sort_pair(block_dict.pop(block_key), index)
                sample.append(pair)
                subsample_size -= 1

                if subsample_size :
                    break
                else :
                    return sample

    else :
        return sample

def linkSamplePredicates(sample_size, predicates, items1, items2) :
    n_1 = len(items1)
    n_2 = len(items2)

    for subsample_size, predicate in subsample(sample_size, predicates) :
        
        if not subsample_size :
            yield None
            continue

        items1.rotate(random.randrange(n_1))
        items2.rotate(random.randrange(n_2))

        try :
            items1.reverse()
            items2.reverse()
        except AttributeError :
            items1 = deque(reversed(items1))
            items2 = deque(reversed(items2))

        yield linkSamplePredicate(subsample_size, predicate, items1, items2)


def linkSamplePredicate(subsample_size, predicate, items1, items2) :
    sample = []

    predicate_function = predicate.func
    field = predicate.field

    red = defaultdict(list)
    blue = defaultdict(list)

    for i, (index, record) in enumerate(interleave(items1, items2)):
        if i == 20000:
            if min(len(red), len(blue)) + len(sample) < 10 :
                return sample

        column = record[field]
        if not column :
            red, blue = blue, red
            continue

        block_keys = predicate_function(column)
        for block_key in block_keys:
            if blue.get(block_key):
                pair = sort_pair(blue[block_key].pop(), index)
                sample.append(pair)

                subsample_size -= 1
                if subsample_size :
                    break
                else :
                    return sample
            else:
                red[block_key].append(index)

        red, blue = blue, red

    for index, record in itertools.islice(items2, len(items1)) :
        column = record[field]
        if not column :
            continue

        block_keys = predicate_function(column)
        for block_key in block_keys:
            if red.get(block_key):
                pair = sort_pair(red[block_key].pop(), index)
                sample.append(pair)

                subsample_size -= 1
                if subsample_size :
                    break
                else :
                    return sample

    return sample

def evenSplits(total_size, num_splits) :
    avg = total_size/num_splits
    split = 0
    for _ in range(num_splits) :
        split += avg - int(split)
        yield int(split)

def subsample(total_size, predicates) :
    splits = evenSplits(total_size, len(predicates))
    for split, predicate in zip(splits, predicates) :
        yield split, predicate

def interleave(*iterables) :
    return itertools.chain.from_iterable(zip(*iterables))

def sort_pair(a, b) :
    if a > b :
        return (b, a)
    else :
        return (a, b)

def randomDeque(data) :
    data_q = deque(random.sample(viewitems(data), len(data)))
    
    return data_q


dedupeBlockedSample = functools.partial(blockedSample, dedupeSamplePredicates) 
linkBlockedSample = functools.partial(blockedSample, linkSamplePredicates) 





#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import division
from builtins import range

import numpy
import logging
import warnings

import dedupe.core

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def gridSearch(training_data,
               learner,
               original_data_model,
               num_cores,
               k=3,
               search_space=[.00001, .0001, .001, .01, .1, 1],
               randomize=True):

    if num_cores < 2 :
        from multiprocessing.dummy import Pool
    else :
        from .backport import Pool

    pool = Pool()

    training_data = training_data[numpy.random.permutation(training_data.size)]

    logger.info('using cross validation to find optimum alpha...')
    best_score = 0
    best_alpha = 0.01

    alpha_tester = AlphaTester(original_data_model, learner)

    for alpha in search_space:

        score_jobs = [pool.apply_async(alpha_tester, 
                                       (training, validation, alpha))
                      for training, validation in 
                      kFolds(training_data, k)]

        scores = [job.get() for job in score_jobs]
        
        average_score = reduceScores(scores)

        logger.debug("Average Score: %f, alpha: %s" % (average_score, alpha))

        if average_score >= best_score :
            best_score = average_score
            best_alpha = alpha

    logger.info('optimum alpha: %f' % best_alpha)
    pool.close()
    pool.join()


    return best_alpha

# http://code.activestate.com/recipes/521906-k-fold-cross-validation-partition/
def kFolds(training_data, k):
    if k < 2 :
        raise ValueError("Number of folds must be at least 2")
    
    if len(training_data) < 2 :
        raise ValueError("At least two traning datum are required")

    train_dtype = training_data.dtype
    slices = [training_data[i::k] for i in range(k)]
    for i in range(k):
        validation = slices[i]
        training = [datum for s in slices if s is not validation for datum in s]
        validation = numpy.array(validation, train_dtype)
        training = numpy.array(training, train_dtype)

        if len(training) and len(validation) :
            yield (training, validation)
        else :
            warnings.warn("Only providing %s folds out of %s requested" % 
                          (i, k))
            break

class AlphaTester(object) :
    def __init__(self, data_model, learner) : # pragma : no cover
        self.data_model = data_model
        self.learner = learner

    def __call__(self, training, validation, alpha) :
        data_model = dedupe.core.trainModel(training, 
                                            self.data_model, 
                                            self.learner, 
                                            alpha)

        weight = numpy.array([field.weight
                              for field in 
                              data_model['fields']])
        bias = data_model['bias']

        predictions = numpy.dot(validation['distances'], weight) + bias
        true_labels = validation['label'] == b'match'

        return scorePredictions(true_labels, predictions)
        
def scorePredictions(true_labels, predictions) :
    # http://en.wikipedia.org/wiki/Matthews_correlation_coefficient

    true_dupes = numpy.sum(predictions[true_labels == 1] > 0)
    false_dupes = numpy.sum(predictions[true_labels == 0] > 0)

    true_distinct = numpy.sum(predictions[true_labels == 0] < 0)
    false_distinct = numpy.sum(predictions[true_labels == 1] < 0)

    if not (true_dupes + false_dupes) * (true_distinct + false_distinct) :
        return 0
    
    matthews_cc = ((true_dupes * true_distinct 
                    - false_dupes * false_distinct)
                   /numpy.sqrt((true_dupes + false_dupes)
                               * (true_dupes + false_distinct)
                               * (true_distinct + false_dupes)
                               * (true_distinct + false_distinct)))


    return matthews_cc

def reduceScores(scores) :
    
    scores = [score for score in scores if score is not None]

    if scores :
        average_score = sum(scores)/len(scores)
    else :
        average_score = 0

    return average_score





from .base import Variable
import itertools

class InteractionType(Variable) :
    type = "Interaction"
    
    def __init__(self, definition) :

        self.interactions = definition["interaction variables"]

        self.name = "(Interaction: %s)" % str(self.interactions)
        self.interaction_fields = self.interactions

        super(InteractionType, self).__init__(definition)

    def expandInteractions(self, field_model) :

        self.interaction_fields = self.atomicInteractions(self.interactions,
                                                          field_model)
        for field in self.interaction_fields :
            if field_model[field].has_missing :
                self.has_missing = True

        self.categorical(field_model)
    
    def categorical(self, field_model) :
        categoricals = [field for field in self.interaction_fields
                        if hasattr(field_model[field], "higher_vars")]
        noncategoricals = [field for field in self.interaction_fields
                           if not hasattr(field_model[field], "higher_vars")]

        dummies = [field_model[field].higher_vars 
                   for field in categoricals]

        self.higher_vars = []
        for combo in itertools.product(*dummies) :
            var_names = [field.name for field in combo] + noncategoricals
            higher_var = InteractionType({'has missing' : self.has_missing,
                                          'interaction variables' : var_names})
            self.higher_vars.append(higher_var)

    def atomicInteractions(self, interactions, field_model) :
        atomic_interactions = []

        for field in interactions :
            try :
                field_definition = field_model[field]
            except KeyError :
                raise KeyError("The interaction variable %s is "\
                               "not a named variable in the variable "\
                               "definition" % field)
            if hasattr(field_model[field], 'interaction_fields') :
                sub_interactions = field_model[field].interaction_fields
                atomic_interactions.extend(self.atomicInteractions(sub_interactions,
                                                                   field_model))
            else :
                atomic_interactions.append(field)

        return atomic_interactions





import numpy
from dedupe import predicates
from .base import FieldType

class PriceType(FieldType) :
    _predicate_functions = [predicates.orderOfMagnitude,
                            predicates.wholeFieldPredicate,
                            predicates.roundTo1]
    type = "Price"

    @staticmethod
    def comparator(price_1, price_2) :
        if price_1 <= 0 :
            return numpy.nan
        elif price_2 <= 0 :
            return numpy.nan
        else :
            return abs(numpy.log10(price_1) - numpy.log10(price_2))



#!/usr/bin/python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup, Extension
except ImportError :
    raise ImportError("setuptools module required, please go to https://pypi.python.org/pypi/setuptools and follow the instructions for installing setuptools")

install_requires=['numpy>=1.9', 
                  'fastcluster', 
                  'dedupe-hcluster',
                  'categorical-distance',
                  'future',
                  'rlr',
                  'metafone',
                  'affinegap',
                  'canonicalize',
                  'simplecosine',
                  'haversine',
                  'BTrees==4.0.8',
                  'simplejson',
                  'zope.interface', 
                  'zope.index']

setup(
    name='dedupe',
    url='https://github.com/datamade/dedupe',
    version='0.8.0.1.3',
    description='A python library for accurate and scaleable data deduplication and entity-resolution',
    packages=['dedupe', 'dedupe.variables'],
    ext_modules=[Extension('dedupe.cpredicates', ['src/cpredicates.c'])],
    license='The MIT License: http://www.opensource.org/licenses/mit-license.php',
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Cython', 
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis'],
    long_description="""
    *dedupe is a library that uses machine learning to perform de-duplication and entity resolution quickly on structured data.*

    **dedupe** will help you: 

    * **remove duplicate entries** from a spreadsheet of names and addresses
    * **link a list** with customer information to another with order history, even without unique customer id's
    * take a database of campaign contributions and **figure out which ones were made by the same person**, even if the names were entered slightly differently for each record

    dedupe takes in human training data and comes up with the best rules for your dataset to quickly and automatically find similar records, even with very large databases.
    
    Important links:
    
    * Documentation: http://dedupe.rtfd.org/
    * Repository: https://github.com/datamade/dedupe
    * Issues: https://github.com/datamade/dedupe/issues
    * Examples: https://github.com/datamade/dedupe-examples
    """
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
dedupe provides the main user interface for the library the
Dedupe class
"""
from __future__ import print_function, division
from future.utils import viewitems, viewvalues

import itertools
import logging
import pickle
import numpy
import multiprocessing
import random
import warnings
import copy
import os
from collections import defaultdict, OrderedDict
import simplejson as json
import rlr

import dedupe
import dedupe.sampling as sampling
import dedupe.core as core
import dedupe.training as training
import dedupe.serializer as serializer
import dedupe.crossvalidation as crossvalidation
import dedupe.predicates as predicates
import dedupe.blocking as blocking
import dedupe.clustering as clustering
from dedupe.datamodel import DataModel

logger = logging.getLogger(__name__)


class Matching(object):
    """
    Base Class for Record Matching Classes
    
    Public methods:

    - `__init__`
    - `thresholdBlocks`
    - `matchBlocks`
    """
    def __init__(self) :
        pass

    def thresholdBlocks(self, blocks, recall_weight=1.5): # pragma : nocover
        """
        Returns the threshold that maximizes the expected F score,
        a weighted average of precision and recall for a sample of
        blocked data. 

        Keyword arguments:
        blocks --        Sequence of tuples of records, where each
                         tuple is a set of records covered by a blocking
                         predicate

        recall_weight -- Sets the tradeoff between precision and
                         recall. I.e. if you care twice as much about
                         recall as you do precision, set recall_weight
                         to 2.
        """


        probability = core.scoreDuplicates(self._blockedPairs(blocks), 
                                           self.data_model, 
                                           self.num_cores)['score']

        probability.sort()
        probability = probability[::-1]

        expected_dupes = numpy.cumsum(probability)

        recall = expected_dupes / expected_dupes[-1]
        precision = expected_dupes / numpy.arange(1, len(expected_dupes) + 1)

        score = recall * precision / (recall + recall_weight ** 2 * precision)

        i = numpy.argmax(score)

        logger.info('Maximum expected recall and precision')
        logger.info('recall: %2.3f', recall[i])
        logger.info('precision: %2.3f', precision[i])
        logger.info('With threshold: %2.3f', probability[i])

        return probability[i]

    def matchBlocks(self, blocks, threshold=.5, *args, **kwargs): # pragma : no cover
        """
        Partitions blocked data and returns a list of clusters, where
        each cluster is a tuple of record ids

        Keyword arguments:
        blocks --     Sequence of tuples of records, where each
                      tuple is a set of records covered by a blocking
                      predicate
                                          
        threshold --  Number between 0 and 1 (default is .5). We will
                      only consider as duplicates record pairs as
                      duplicates if their estimated duplicate likelihood is
                      greater than the threshold.

                      Lowering the number will increase recall, raising it
                      will increase precision
                              

        """
        # Setting the cluster threshold this ways is not principled,
        # but seems to reliably help performance
        cluster_threshold = threshold * 0.7

        candidate_records = self._blockedPairs(blocks)
        
        matches = core.scoreDuplicates(candidate_records,
                                       self.data_model,
                                       self.num_cores,
                                       threshold)

        logger.debug("matching done, begin clustering")

        clusters = self._cluster(matches, 
                                 cluster_threshold, *args, **kwargs)

        try :
            match_file = matches.filename
            del matches
            os.remove(match_file)
        except AttributeError :
            pass
        
        return clusters

    def _checkRecordType(self, record) :
        for field_comparator in self.data_model.field_comparators :
            field = field_comparator[0]
            if field not in record :
                raise ValueError("Records do not line up with data model. "
                                 "The field '%s' is in data_model but not "
                                 "in a record" % field)


    def _logLearnedWeights(self): # pragma: no cover
        """
        Log learned weights and bias terms
        """
        logger.info('Learned Weights')
        for (key_1, value_1) in self.data_model.items():
            try:
                for field in value_1 :
                    logger.info((field.name, field.weight))
            except TypeError :
                logger.info((key_1, value_1))


class DedupeMatching(Matching) :
    """
    Class for Deduplication, extends Matching.
    
    Use DedupeMatching when you have a dataset that can contain 
    multiple references to the same entity.
    
    Public methods:

    - `__init__`
    - `match`
    - `threshold`
    """
    
    def __init__(self, *args, **kwargs) :
        super(DedupeMatching, self).__init__(*args, **kwargs)
        self._cluster = clustering.cluster
        self._linkage_type = "Dedupe"

    def match(self, data, threshold = 0.5) : # pragma : no cover
        """
        Identifies records that all refer to the same entity, returns tuples
        containing a set of record ids and a confidence score as a float between 0
        and 1. The record_ids within each set should refer to the
        same entity and the confidence score is a measure of our confidence that
        all the records in a cluster refer to the same entity.
        
        This method should only used for small to moderately sized datasets
        for larger data, use matchBlocks
        
        Arguments:
        data      --  Dictionary of records, where the keys are record_ids
                      and the values are dictionaries with the keys being
                      field names
                                          
        threshold --  Number between 0 and 1 (default is .5). We will consider
                      records as potential duplicates if the predicted probability
                      of being a duplicate is above the threshold.

                      Lowering the number will increase recall, raising it
                      will increase precision
                             
        """
        blocked_pairs = self._blockData(data)
        return self.matchBlocks(blocked_pairs, threshold)

    def threshold(self, data, recall_weight = 1.5) : # pragma : no cover
        """
        Returns the threshold that maximizes the expected F score,
        a weighted average of precision and recall for a sample of
        data. 

        Arguments:
        data          -- Dictionary of records, where the keys are record_ids
                         and the values are dictionaries with the keys being
                         field names

        recall_weight -- Sets the tradeoff between precision and
                         recall. I.e. if you care twice as much about
                         recall as you do precision, set recall_weight
                         to 2.
        """

    
        blocked_pairs = self._blockData(data)
        return self.thresholdBlocks(blocked_pairs, recall_weight)

    def _blockedPairs(self, blocks) :
        """
        Generate tuples of pairs of records from a block of records
        
        Arguments:
        
        blocks -- an iterable sequence of blocked records
        """
        
        block, blocks = core.peek(blocks)
        self._checkBlock(block)

        combinations = itertools.combinations

        pairs = (combinations(block, 2) for block in blocks)

        return itertools.chain.from_iterable(pairs) 

    def _checkBlock(self, block) :
        if block :
            try :
                if len(block[0]) < 3 :
                    raise ValueError("Each item in a block must be a "
                                     "sequence of record_id, record, and smaller ids "
                                     "and the records also must be dictionaries")
            except :
                raise ValueError("Each item in a block must be a "
                                 "sequence of record_id, record, and smaller ids "
                                 "and the records also must be dictionaries")
            try :
                block[0][1].items()
                block[0][2].isdisjoint([])
            except :
                raise ValueError("The record must be a dictionary and "
                                 "smaller_ids must be a set")

        
            self._checkRecordType(block[0][1])

    def _blockData(self, data_d):

        blocks = defaultdict(dict)

        for field in self.blocker.index_fields :
            unique_fields = {record[field]
                             for record 
                             in viewvalues(data_d)
                             if record[field]}

            self.blocker.index(unique_fields, field)

        for block_key, record_id in self.blocker(viewitems(data_d)) :
            blocks[block_key][record_id] = data_d[record_id]

        self.blocker.resetIndices()

        blocks = (records for records in blocks.values()
                  if len(records) > 1)
        
        blocks = {frozenset(d.keys()) : d for d in blocks}
        blocks = blocks.values()

        for block in self._redundantFree(blocks) :
            yield block

    def _redundantFree(self, blocks) :
        """
        Redundant-free Comparisons from Kolb et al, "Dedoop:
        Efficient Deduplication with Hadoop"
        http://dbs.uni-leipzig.de/file/Dedoop.pdf
        """
        coverage = defaultdict(list)

        for block_id, records in enumerate(blocks) :

            for record_id, record in viewitems(records) :
                coverage[record_id].append(block_id)

        for block_id, records in enumerate(blocks) :
            if block_id % 10000 == 0 :
                logger.info("%s blocks" % block_id)

            marked_records = []
            for record_id, record in viewitems(records) :
                smaller_ids = {covered_id for covered_id 
                               in coverage[record_id] 
                               if covered_id < block_id}
                marked_records.append((record_id, record, smaller_ids))

            yield marked_records


class RecordLinkMatching(Matching) :
    """
    Class for Record Linkage, extends Matching.
    
    Use RecordLinkMatching when you have two datasets that you want to merge
    where each dataset, individually, contains no duplicates.
    
    Public methods:

    - `__init__`
    - `match`
    - `threshold`
    """

    def __init__(self, *args, **kwargs) :
        super(RecordLinkMatching, self).__init__(*args, **kwargs)

        self._cluster = clustering.greedyMatching
        self._linkage_type = "RecordLink"

    def match(self, data_1, data_2, threshold = 1.5) : # pragma : no cover
        """
        Identifies pairs of records that refer to the same entity, returns
        tuples containing a set of record ids and a confidence score as a float
        between 0 and 1. The record_ids within each set should refer to the 
        same entity and the confidence score is the estimated probability that
        the records refer to the same entity.
        
        This method should only used for small to moderately sized datasets
        for larger data, use matchBlocks
        
        Arguments:
        data_1    -- Dictionary of records from first dataset, where the 
                     keys are record_ids and the values are dictionaries
                     with the keys being field names

        data_2    -- Dictionary of records from second dataset, same form 
                     as data_1
                                          
        threshold -- Number between 0 and 1 (default is .5). We will consider
                     records as potential duplicates if the predicted 
                     probability of being a duplicate is above the threshold.

                     Lowering the number will increase recall, raising it
                     will increase precision
        """

        blocked_pairs = self._blockData(data_1, data_2)
        return self.matchBlocks(blocked_pairs, threshold)

    def threshold(self, data_1, data_2, recall_weight = 1.5) : # pragma : no cover
        """
        Returns the threshold that maximizes the expected F score,
        a weighted average of precision and recall for a sample of
        data. 

        Arguments:
        data_1        --  Dictionary of records from first dataset, where the 
                          keys are record_ids and the values are dictionaries 
                          with the keys being field names

        data_2        --  Dictionary of records from second dataset, same form 
                          as data_1

        recall_weight -- Sets the tradeoff between precision and
                         recall. I.e. if you care twice as much about
                         recall as you do precision, set recall_weight
                         to 2.
x        """

        blocked_pairs = self._blockData(data_1, data_2)
        return self.thresholdBlocks(blocked_pairs, recall_weight)

    def _blockedPairs(self, blocks) :
        """
        Generate tuples of pairs of records from a block of records
        
        Arguments:
        
        blocks -- an iterable sequence of blocked records
        """
        
        block, blocks = core.peek(blocks)
        self._checkBlock(block)

        product = itertools.product

        pairs = (product(base, target) for base, target in blocks)

        return itertools.chain.from_iterable(pairs) 
        
    def _checkBlock(self, block) :
        if block :
            try :
                base, target = block
            except :
                raise ValueError("Each block must be a made up of two "
                                 "sequences, (base_sequence, target_sequence)")

            if base :
                if len(base[0]) < 3 :
                    raise ValueError("Each sequence must be made up of 3-tuple "
                                     "like (record_id, record, covered_blocks)")
                self._checkRecordType(base[0][1])
            if target :
                if len(target[0]) < 3 :
                    raise ValueError("Each sequence must be made up of 3-tuple "
                                     "like (record_id, record, covered_blocks)")
                self._checkRecordType(target[0][1])

    def _blockGenerator(self, messy_data, blocked_records) :
        block_groups = itertools.groupby(self.blocker(viewitems(messy_data)), 
                                         lambda x : x[1])

        for i, (record_id, block_keys) in enumerate(block_groups) :
            if i % 100 == 0 :
                logger.info("%s records" % i)

            A = [(record_id, messy_data[record_id], set())]

            B = {}

            for block_key, _ in block_keys :
                if block_key in blocked_records :
                    B.update(blocked_records[block_key])

            B = [(rec_id, record, set())
                 for rec_id, record
                 in B.items()]

            if B :
                yield (A, B)


    def _blockData(self, data_1, data_2) :

        blocked_records = defaultdict(dict)

        for field in self.blocker.index_fields :
            fields_2 = (record[field]
                        for record 
                        in viewvalues(data_2))

            self.blocker.index(set(fields_2), field)

        for block_key, record_id in self.blocker(data_2.items()) :
            blocked_records[block_key][record_id] = data_2[record_id]

        for each in self._blockGenerator(data_1, blocked_records) :
            yield each

        self.blocker.resetIndices()


class StaticMatching(Matching) :
    """
    Class for initializing a dedupe object from a settings file, extends Matching.
    
    Public methods:
    - __init__
    """

    def __init__(self, 
                 settings_file, 
                 num_cores=None) : # pragma : no cover
        """
        Initialize from a settings file
        #### Example usage

            # initialize from a settings file
            with open('my_learned_settings', 'rb') as f:
                deduper = dedupe.StaticDedupe(f)

        #### Keyword arguments
        
        `settings_file`
        A file object containing settings data.


        Settings files are typically generated by saving the settings
        learned from ActiveMatching. If you need details for this
        file see the method [`writeSettings`][[api.py#writesettings]].
        """
        if num_cores is None :
            self.num_cores = multiprocessing.cpu_count()
        else :
            self.num_cores = num_cores

        try:
            self.data_model = pickle.load(settings_file)
            self.predicates = pickle.load(settings_file)
            self.stop_words = pickle.load(settings_file)
        except (KeyError, AttributeError) :
            raise ValueError("This settings file is not compatible with "
                             "the current version of dedupe. This can happen "
                             "if you have recently upgraded dedupe.")
        except :
            print("Something has gone wrong with loading the settings file")
            raise
                             

        self._logLearnedWeights()
        logger.info(self.predicates)
        logger.info(self.stop_words)

        self.blocker = blocking.Blocker(self.predicates, 
                                        self.stop_words)





class ActiveMatching(Matching) :
    """
    Class for training dedupe extends Matching.
    
    Public methods:
    - __init__
    - readTraining
    - train
    - writeSettings
    - writeTraining
    - uncertainPairs
    - markPairs
    - cleanupTraining
    """

    def __init__(self, 
                 variable_definition, 
                 data_sample = None,
                 num_cores = None) :
        """
        Initialize from a data model and data sample.

        #### Example usage

            # initialize from a defined set of fields
            fields = [{'field' : 'Site name', 'type': 'String'},
                      {'field' : 'Address', 'type': 'String'},
                      {'field' : 'Zip', 'type': 'String', 'Has Missing':True},
                      {'field' : 'Phone', 'type': 'String', 'Has Missing':True},
                     ]

            data_sample = [
                           (
                            (854, {'city': 'san francisco',
                             'address': '300 de haro st.',
                             'name': "sally's cafe & bakery",
                             'cuisine': 'american'}),
                            (855, {'city': 'san francisco',
                             'address': '1328 18th st.',
                             'name': 'san francisco bbq',
                             'cuisine': 'thai'})
                             )
                            ]


            
            deduper = dedupe.Dedupe(fields, data_sample)

        
        #### Additional detail

        A field definition is a list of dictionaries where each dictionary
        describes a variable to use for comparing records. 

        For details about variable types, check the documentation.
        <http://dedupe.readthedocs.org>`_ 

        In the data_sample, each element is a tuple of two
        records. Each record is, in turn, a tuple of the record's key and
        a record dictionary.

        In in the record dictionary the keys are the names of the
        record field and values are the record values.
        """
        self.data_model = DataModel(variable_definition)

        self.data_sample = data_sample

        if self.data_sample :
            self._checkDataSample(self.data_sample)
            self.activeLearner = training.ActiveLearning(self.data_sample, 
                                                         self.data_model)
        else :
            self.data_sample = []
            self.activeLearner = None

        if num_cores is None :
            self.num_cores = multiprocessing.cpu_count()
        else :
            self.num_cores = num_cores

        training_dtype = [('label', 'S8'), 
                          ('distances', 'f4', 
                           (len(self.data_model['fields']), ))]

        self.training_data = numpy.zeros(0, dtype=training_dtype)
        self.training_pairs = OrderedDict({u'distinct': [], 
                                           u'match': []})

        self.learner = rlr.lr
        self.blocker = None


    def cleanupTraining(self) : # pragma : no cover
        '''
        Clean up data we used for training. Free up memory.
        '''
        del self.training_data
        del self.training_pairs
        del self.activeLearner
        del self.data_sample


    def readTraining(self, training_file) :
        '''
        Read training from previously built training data file object
        
        Arguments:
        
        training_file -- file object containing the training data
        '''
        
        logger.info('reading training from file')
        
        training_pairs = json.load(training_file, 
                                   cls=serializer.dedupe_decoder)


        for (label, examples) in training_pairs.items():
            if examples :
                self._checkRecordPairType(examples[0])

            examples = core.freezeData(examples)

            training_pairs[label] = examples
            self.training_pairs[label].extend(examples)

        self._addTrainingData(training_pairs)

        self._trainClassifier(0.1)

    def train(self, ppc=.1, uncovered_dupes=1, index_predicates=True) : # pragma : no cover
        """Keyword arguments:
        ppc -- Limits the Proportion of Pairs Covered that we allow a
               predicate to cover. If a predicate puts together a fraction
               of possible pairs greater than the ppc, that predicate will
               be removed from consideration.

               As the size of the data increases, the user will generally
               want to reduce ppc.

               ppc should be a value between 0.0 and 1.0

        uncovered_dupes -- The number of true dupes pairs in our training
                           data that we can accept will not be put into any
                           block. If true true duplicates are never in the
                           same block, we will never compare them, and may
                           never declare them to be duplicates.

                           However, requiring that we cover every single
                           true dupe pair may mean that we have to use
                           blocks that put together many, many distinct pairs
                           that we'll have to expensively, compare as well.

        index_predicates -- Should dedupe consider predicates that
                            rely upon indexing the data. Index predicates can 
                            be slower and take susbstantial memory.

                            Defaults to True.

        """
        self._trainClassifier()
        self._trainBlocker(ppc, uncovered_dupes, index_predicates)

    def _trainClassifier(self, alpha=None) : # pragma : no cover

        if alpha is None :
            alpha = self._regularizer()

        self.data_model = core.trainModel(self.training_data,
                                          self.data_model, 
                                          self.learner,
                                          alpha)

        self._logLearnedWeights()

    def _regularizer(self) :
        n_folds = min(numpy.sum(self.training_data['label']==u'match')/3,
                      20)
        n_folds = max(n_folds,
                      2)

        logger.info('%d folds', n_folds)

        alpha = crossvalidation.gridSearch(self.training_data,
                                           self.learner,
                                           self.data_model, 
                                           self.num_cores,
                                           k=n_folds)

        return alpha


    
    def _trainBlocker(self, ppc=1, uncovered_dupes=1, index_predicates=True) : # pragma : no cover
        training_pairs = copy.deepcopy(self.training_pairs)

        confident_nonduplicates = training.semiSupervisedNonDuplicates(self.data_sample,
                                                                       self.data_model,
                                                                       sample_size=32000)

        training_pairs[u'distinct'].extend(confident_nonduplicates)

        predicate_set = predicateGenerator(self.data_model, 
                                           index_predicates,
                                           self.canopies)

        (self.predicates, 
         self.stop_words) = dedupe.training.blockTraining(training_pairs,
                                                          predicate_set,
                                                          ppc,
                                                          uncovered_dupes,
                                                          self._linkage_type)

        self.blocker = blocking.Blocker(self.predicates,
                                        self.stop_words) 


    def writeSettings(self, file_obj): # pragma : no cover
        """
        Write a settings file containing the 
        data model and predicates to a file object

        Keyword arguments:
        file_obj -- file object to write settings data into
        """

        pickle.dump(self.data_model, file_obj)
        pickle.dump(self.predicates, file_obj)
        pickle.dump(dict(self.stop_words), file_obj)

    def writeTraining(self, file_obj): # pragma : no cover
        """
        Write to a json file that contains labeled examples

        Keyword arguments:
        file_obj -- file object to write training data to
        """

        json.dump(self.training_pairs, 
                  file_obj, 
                  cls=serializer.dedupe_encoder,
                  ensure_ascii=True)


    def uncertainPairs(self) :
        '''
        Provides a list of the pairs of records that dedupe is most
        curious to learn if they are matches or distinct.
        
        Useful for user labeling.

        '''
        
        
        if self.training_data.shape[0] == 0 :
            rand_int = random.randint(0, len(self.data_sample)-1)
            random_pair = self.data_sample[rand_int]
            exact_match = (random_pair[0], random_pair[0]) 
            self._addTrainingData({u'match':[exact_match, exact_match],
                                   u'distinct':[]})


        self._trainClassifier(0.1)

        
        dupe_ratio = (len(self.training_pairs[u'match'])
                      /(len(self.training_pairs[u'distinct']) + 1))

        return self.activeLearner.uncertainPairs(self.data_model, dupe_ratio)

    def markPairs(self, labeled_pairs) :
        '''
        Add a labeled pairs of record to dedupes training set and update the
        matching model
        
        Argument :

        labeled_pairs -- A dictionary with two keys, `match` and `distinct`
                         the values are lists that can contain pairs of records
                         
        '''
        try :
            labeled_pairs.items()
            labeled_pairs[u'match']
            labeled_pairs[u'distinct']
        except :
            raise ValueError('labeled_pairs must be a dictionary with keys '
                             '"distinct" and "match"')

        if labeled_pairs[u'match'] :
            pair = labeled_pairs[u'match'][0]
            self._checkRecordPairType(pair)
        
        if labeled_pairs[u'distinct'] :
            pair = labeled_pairs[u'distinct'][0]
            self._checkRecordPairType(pair)
        
        if not labeled_pairs[u'distinct'] and not labeled_pairs[u'match'] :
            warnings.warn("Didn't return any labeled record pairs")
        

        for label, pairs in labeled_pairs.items() :
            self.training_pairs[label].extend(core.freezeData(pairs))

        self._addTrainingData(labeled_pairs) 



    def _checkRecordPairType(self, record_pair) :
        try :
            record_pair[0]
        except :
            raise ValueError("The elements of data_sample must be pairs "
                             "of record_pairs (ordered sequences of length 2)")

        if len(record_pair) != 2 :
            raise ValueError("The elements of data_sample must be pairs "
                             "of record_pairs")
        try :
            record_pair[0].keys() and record_pair[1].keys()
        except :
            raise ValueError("A pair of record_pairs must be made up of two "
                             "dictionaries ")

        self._checkRecordType(record_pair[0])
        self._checkRecordType(record_pair[1])

    def  _checkDataSample(self, data_sample) :
        try :
            len(data_sample)
        except TypeError :
            raise ValueError("data_sample must be a sequence")

        if len(data_sample) :
            self._checkRecordPairType(data_sample[0])

        else :
            warnings.warn("You submitted an empty data_sample")




    def _addTrainingData(self, labeled_pairs) :
        """
        Appends training data to the training data collection.
        """
    
        for label, examples in labeled_pairs.items () :
            n_examples = len(examples)
            labels = [label] * n_examples

            new_data = numpy.empty(n_examples,
                                   dtype=self.training_data.dtype)

            new_data['label'] = labels
            new_data['distances'] = core.fieldDistances(examples, 
                                                        self.data_model)

            self.training_data = numpy.append(self.training_data, 
                                              new_data)


    def _loadSample(self, data_sample) :

        self._checkDataSample(data_sample) 

        self.data_sample = data_sample

        self.activeLearner = training.ActiveLearning(self.data_sample, 
                                                     self.data_model)



class StaticDedupe(DedupeMatching, StaticMatching) :
    """
    Mixin Class for Static Deduplication
    """

class Dedupe(DedupeMatching, ActiveMatching) :
    """
    Mixin Class for Active Learning Deduplication
    
    Public Methods
    - sample
    """
    canopies = True

    def sample(self, data, sample_size=15000, 
               blocked_proportion=0.5) :
        '''Draw a sample of record pairs from the dataset
        (a mix of random pairs & pairs of similar records)
        and initialize active learning with this sample
        
        Arguments: data -- Dictionary of records, where the keys are
        record_ids and the values are dictionaries with the keys being
        field names
        
        sample_size         -- Size of the sample to draw
        blocked_proportion  -- Proportion of the sample that will be blocked
        '''
        data = core.index(data)

        blocked_sample_size = int(blocked_proportion * sample_size)
        predicates = list(predicateGenerator(self.data_model, 
                                             index_predicates=False,
                                             canopies=self.canopies))


        data = sampling.randomDeque(data)
        blocked_sample_keys = sampling.dedupeBlockedSample(blocked_sample_size,
                                                           predicates,
                                                           data)

        random_sample_size = sample_size - len(blocked_sample_keys)
        random_sample_keys = set(dedupe.core.randomPairs(len(data),
                                                         random_sample_size))
        data = dict(data)

        data_sample = ((data[k1], data[k2])
                       for k1, k2 
                       in blocked_sample_keys | random_sample_keys)

        data_sample = core.freezeData(data_sample)

        self._loadSample(data_sample)


class StaticRecordLink(RecordLinkMatching, StaticMatching) :
    """
    Mixin Class for Static Record Linkage
    """

class RecordLink(RecordLinkMatching, ActiveMatching) :
    """
    Mixin Class for Active Learning Record Linkage
    
    Public Methods
    - sample
    """
    canopies = False

    def sample(self, data_1, data_2, sample_size=150000, 
               blocked_proportion=.5) :
        '''
        Draws a random sample of combinations of records from 
        the first and second datasets, and initializes active
        learning with this sample
        
        Arguments:
        
        data_1      -- Dictionary of records from first dataset, where the 
                       keys are record_ids and the values are dictionaries 
                       with the keys being field names
        data_2      -- Dictionary of records from second dataset, same 
                       form as data_1
        
        sample_size -- Size of the sample to draw
        '''
        if len(data_1) > len(data_2) :
            data_1, data_2 = data_2, data_1

        data_1 = core.index(data_1)

        offset = len(data_1)
        data_2 = core.index(data_2, offset)

        blocked_sample_size = int(blocked_proportion * sample_size)
        predicates = list(predicateGenerator(self.data_model, 
                                             index_predicates=False,
                                             canopies=self.canopies))

        data_1 = sampling.randomDeque(data_1)
        data_2 = sampling.randomDeque(data_2)

        blocked_sample_keys = sampling.linkBlockedSample(blocked_sample_size,
                                                         predicates,
                                                         data_1, 
                                                         data_2)
        
        random_sample_size = sample_size - len(blocked_sample_keys)
        random_sample_keys = dedupe.core.randomPairsMatch(len(data_1),
                                                          len(data_2), 
                                                          random_sample_size)

        random_sample_keys = {(a, b + offset) 
                              for a, b in random_sample_keys}

        data_1 = dict(data_1)
        data_2 = dict(data_2)
        
        data_sample = ((data_1[k1], data_2[k2])
                       for k1, k2 
                       in blocked_sample_keys | random_sample_keys)

        data_sample = core.freezeData(data_sample)

        self._loadSample(data_sample)


class GazetteerMatching(RecordLinkMatching) :
    
    def __init__(self, *args, **kwargs) :
        super(GazetteerMatching, self).__init__(*args, **kwargs)

        self._cluster = clustering.gazetteMatching
        self._linkage_type = "GazetteerMatching"
        self.blocked_records = OrderedDict({})


    def _blockData(self, messy_data) :
        for each in self._blockGenerator(messy_data, self.blocked_records) :
            yield each


    def index(self, data) : # pragma : no cover

        for field in self.blocker.index_fields :
            self.blocker.index((record[field]
                                for record 
                                in viewvalues(data)),
                               field)

        for block_key, record_id in self.blocker(data.items()) :
            if block_key not in self.blocked_records :
                self.blocked_records[block_key] = {}
            self.blocked_records[block_key][record_id] = data[record_id]

    def unindex(self, data) : # pragma : no cover

        for field in self.blocker.index_fields :
            self.blocker.unindex((record[field]
                                  for record 
                                  in viewvalues(data)),
                                 field)

        for block_key, record_id in self.blocker(viewitems(data)) :
            try : 
                del self.blocked_records[block_key][record_id]
            except KeyError :
                pass 
        
    def match(self, messy_data, threshold = 0.5, n_matches = 1) : # pragma : no cover
        """Identifies pairs of records that refer to the same entity, returns
        tuples containing a set of record ids and a confidence score as a float
        between 0 and 1. The record_ids within each set should refer to the 
        same entity and the confidence score is the estimated probability that
        the records refer to the same entity.
        
        This method should only used for small to moderately sized datasets
        for larger data, use matchBlocks
        
        Arguments:
        messy_data -- Dictionary of records from messy dataset, where the 
                      keys are record_ids and the values are dictionaries with 
                      the keys being field names

        threshold -- Number between 0 and 1 (default is .5). We will consider
                     records as potential duplicates if the predicted 
                     probability of being a duplicate is above the threshold.

                     Lowering the number will increase recall, raising it
                     will increase precision
        
        n_matches -- Maximum number of possible matches from the canonical 
                     record set to match against each record in the messy
                     record set
        """
        blocked_pairs = self._blockData(messy_data)
        return self.matchBlocks(blocked_pairs, threshold, n_matches)

class Gazetteer(RecordLink, GazetteerMatching):
    pass

class StaticGazetteer(StaticRecordLink, GazetteerMatching):
    pass

def predicateGenerator(data_model, index_predicates, canopies) :
    predicates = set()
    for definition in data_model.primary_fields :
        for predicate in definition.predicates :
            if hasattr(predicate, 'index') :
                if index_predicates :
                    if hasattr(predicate, 'canopy') :
                        if canopies :
                            predicates.add(predicate)
                    else :
                        if not canopies :
                            predicates.add(predicate)
            else :
                predicates.add(predicate)

    return predicates




#!/usr/bin/python
# -*- coding: utf-8 -*-
from builtins import range, next, zip
from future.utils import viewvalues
import sys
if sys.version < '3':
    text_type = unicode
    binary_type = str
else:
    text_type = str
    binary_type = bytes
    unicode = str

import itertools
import warnings
import numpy
import collections
import time
import tempfile
import os

import dedupe.backport as backport

class ChildProcessError(Exception) :
    pass

def randomPairsWithReplacement(n_records, sample_size) :
    # If the population is very large relative to the sample
    # size than we'll get very few duplicates by chance
    warnings.warn("There may be duplicates in the sample")

    try :
        random_indices = numpy.random.randint(n_records, 
                                              size=sample_size*2)
    except OverflowError:
        max_int = numpy.iinfo('int').max
        warnings.warn("Asked to sample pairs from %d records, will only sample pairs from first %d records" % (n_records, max_int))
        random_indices = numpy.random.randint(max_int, 
                                              size=sample_size*2)


        
    random_indices = random_indices.reshape((-1, 2))
    random_indices.sort(axis=1)

    return [(p.item(), q.item()) for p, q in random_indices]


def randomPairs(n_records, sample_size):
    """
    Return random combinations of indices for a square matrix of size
    n records
    """

    if n_records < 2 :
        raise ValueError("Needs at least two records")
    n = n_records * (n_records - 1) / 2

    # numpy doesn't always throw an overflow error so we need to 
    # check to make sure that the largest number we'll use is smaller
    # than the numpy's maximum unsigned integer
    if 8 * n > numpy.iinfo('uint').max :
        return randomPairsWithReplacement(n_records, sample_size)

    if sample_size >= n:
        if sample_size > n :
            warnings.warn("Requested sample of size %d, only returning %d possible pairs" % (sample_size, n))

        random_indices = numpy.arange(n)
    else :
        random_indices = numpy.random.randint(int(n), size=sample_size)

    random_indices = random_indices.astype('uint')

    b = 1 - 2 * n_records

    x = numpy.trunc((-b - numpy.sqrt(b ** 2 - 8 * random_indices)) / 2)
    y = random_indices + x * (b + x + 2) / 2 + 1

    stacked = numpy.column_stack((x, y)).astype(int)

    return [(p.item(), q.item()) for p, q in stacked]

def randomPairsMatch(n_records_A, n_records_B, sample_size):
    """
    Return random combinations of indices for record list A and B
    """
    if not n_records_A or not n_records_B :
        raise ValueError("There must be at least one record in A and in B")

    if sample_size >= n_records_A * n_records_B :

        if sample_size > n_records_A * n_records_B :
            warnings.warn("Requested sample of size %d, only returning %d possible pairs" % (sample_size, n_records_A * n_records_B))

        return cartesian((numpy.arange(n_records_A),
                          numpy.arange(n_records_B)))

    A_samples = numpy.random.randint(n_records_A, size=sample_size)
    B_samples = numpy.random.randint(n_records_B, size=sample_size)
    pairs = zip(A_samples,B_samples)
    set_pairs = set(pairs)

    while len(set_pairs) < sample_size:
        set_pairs.update(randomPairsMatch(n_records_A,
                                          n_records_B,
                                          (sample_size-len(set_pairs))))
    else:
        return set_pairs


def trainModel(training_data, data_model, learner=None, alpha=.001):
    """
    Use logistic regression to train weights for all fields in the data model
    """
    
    labels = numpy.array(training_data['label'] == b'match', dtype='i4')
    examples = training_data['distances']

    weight, bias = learner(labels, examples, alpha)

    for i, field_definition in enumerate(data_model['fields']) :
        field_definition.weight = float(weight[i])

    data_model['bias'] = bias

    return data_model

def fieldDistances(record_pairs, data_model):
    num_records = len(record_pairs)

    distances = numpy.empty((num_records, data_model.n_fields))
    field_comparators = data_model.field_comparators

    for i, (record_1, record_2) in enumerate(record_pairs) :
        for field, compare, start, stop in field_comparators :
            distances[i,start:stop] = compare(record_1[field],
                                              record_2[field])
    
    distances = derivedDistances(distances, data_model)

    return distances

def derivedDistances(primary_distances, data_model) :
    distances = primary_distances

    current_column = data_model.derived_start

    for interaction in data_model.interactions :
        distances[:,current_column] =\
                numpy.prod(distances[:,interaction], axis=1)

        current_column += 1

    missing_data = numpy.isnan(distances[:,:current_column])

    distances[:,:current_column][missing_data] = 0

    if data_model.missing_field_indices :
        distances[:,current_column:] =\
            1 - missing_data[:,data_model.missing_field_indices]

    return distances


def scorePairs(field_distances, data_model):
    fields = data_model['fields']

    field_weights = [field.weight for field in fields]
    bias = data_model['bias']

    scores = numpy.dot(field_distances, field_weights)

    scores = numpy.exp(scores + bias) / (1 + numpy.exp(scores + bias))

    return scores

class ScoreRecords(object) :
    def __init__(self, data_model, threshold) :
        self.data_model = data_model
        self.threshold = threshold
        self.score_queue = None

    def __call__(self, records_queue, score_queue) :
        self.score_queue = score_queue
        while True :
            record_pairs = records_queue.get()
            if record_pairs is None :
                break

            try :
                filtered_pairs = self.fieldDistance(record_pairs)
                if filtered_pairs is not None :
                    score_queue.put(filtered_pairs)
            except Exception as e :
                score_queue.put(e)
                raise

        score_queue.put(None)

    def fieldDistance(self, record_pairs) :
        ids = []
        records = []
        
        for record_pair in record_pairs :
            ((id_1, record_1, smaller_ids_1), 
             (id_2, record_2, smaller_ids_2)) = record_pair

            if set.isdisjoint(smaller_ids_1, smaller_ids_2) :
                
                ids.append((id_1, id_2))
                records.append((record_1, record_2))

        if records :
            distances = fieldDistances(records, self.data_model)
            scores = scorePairs(distances, self.data_model)

            scored_pairs = numpy.rec.fromarrays((ids, scores),
                                                dtype= [('pairs', 'object', 2), 
                                                        ('score', 'f4', 1)])
            
            filtered_pairs = scored_pairs[scores > self.threshold]

            return filtered_pairs

def mergeScores(score_queue, result_queue, stop_signals) :
    scored_pairs = numpy.empty(0, dtype= [('pairs', 'object', 2), 
                                          ('score', 'f4', 1)])

    seen_signals = 0
    while seen_signals < stop_signals  :
        score_chunk = score_queue.get()
        if isinstance(score_chunk, Exception) :
            result_queue.put(score_chunk)
            return

        if score_chunk is not None :
            scored_pairs = numpy.concatenate((scored_pairs, score_chunk))
        else :
            seen_signals += 1

    if len(scored_pairs) :
        python_type = type(scored_pairs['pairs'][0][0])
        if python_type is binary_type or python_type is text_type :
            max_length = len(max(numpy.ravel(scored_pairs['pairs']), key=len))
            python_type = (unicode, max_length)
        
        write_dtype = [('pairs', python_type, 2),
                       ('score', 'f4', 1)]

        scored_pairs = scored_pairs.astype(write_dtype)

        scored_pairs_file, file_path = tempfile.mkstemp()
        
        os.close(scored_pairs_file)

        fp = numpy.memmap(file_path, 
                          dtype=scored_pairs.dtype, 
                          shape=scored_pairs.shape)
        fp[:] = scored_pairs[:]

        result_queue.put((file_path, scored_pairs.dtype))

    else :
        result_queue.put(scored_pairs)

def scoreDuplicates(records, data_model, num_cores=1, threshold=0) :
    if num_cores < 2 :
        from multiprocessing.dummy import Process, Pool, Queue
        SimpleQueue = Queue
    else :
        from .backport import Process, Pool, SimpleQueue

    record_pairs_queue = SimpleQueue()
    score_queue =  SimpleQueue()
    result_queue = SimpleQueue()

    n_map_processes = max(num_cores-1, 1)
    score_records = ScoreRecords(data_model, threshold) 
    map_processes = [Process(target=score_records,
                             args=(record_pairs_queue,
                                   score_queue))
                     for _ in range(n_map_processes)]
    [process.start() for process in map_processes]

    reduce_process = Process(target=mergeScores,
                             args=(score_queue,
                                   result_queue,
                                   n_map_processes))
    reduce_process.start()

    fillQueue(record_pairs_queue, records, n_map_processes)

    result = result_queue.get()
    if isinstance(result, Exception) :
        raise ChildProcessError

    if result :
        scored_pairs_file, dtype = result
        scored_pairs = numpy.memmap(scored_pairs_file,
                                    dtype=dtype)
    else :
        scored_pairs = result

    return scored_pairs



def fillQueue(queue, iterable, stop_signals) :
    iterable = iter(iterable)
    chunk_size = 100000
    multiplier = 1.1

    # initial values
    i = 0
    n_records = 0
    t0 = time.clock()
    last_rate = 10000

    while True :
        chunk = list(itertools.islice(iterable, int(chunk_size)))
        if chunk :
            queue.put(chunk)

            n_records += chunk_size
            i += 1

            if i % 10 :
                time_delta = max(time.clock() - t0, 0.0001)

                current_rate = n_records/time_delta

                # chunk_size is always either growing or shrinking, if
                # the shrinking led to a faster rate, keep
                # shrinking. Same with growing. If the rate decreased,
                # reverse directions
                if current_rate < last_rate :
                    multiplier = 1/multiplier

                chunk_size = max(chunk_size * multiplier, 1)

                last_rate = current_rate
                n_records = 0
                t0 = time.clock()
                

        else :
            # put poison pills in queue to tell scorers that they are
            # done
            [queue.put(None) for _ in range(stop_signals)]
            break

def peek(records) :
    try :
        record = next(records)
    except TypeError as e:
        if "not an iterator" not in str(e) :
            raise
        try :
            records = iter(records)
            record = next(records)
        except StopIteration :
            return None, records
    except StopIteration :
        return None, records
    


    return record, itertools.chain([record], records)


def freezeData(data) : # pragma : no cover
    lfrozendict = frozendict
    return [(lfrozendict(record_1), 
             lfrozendict(record_2))
            for record_1, record_2 in data]

def isIndexed(data, offset) :
    hashable = collections.Hashable
    for i in range(offset, offset + len(data)) :
        if i not in data :
            return False
    else :
        return True

def index(data, offset=0) :
    if isIndexed(data, offset) :
        return data
    else :
        data = dict(zip(itertools.count(offset), 
                        viewvalues(data)))
        return data


class frozendict(collections.Mapping):
    """Don't forget the docstrings!!"""

    def __init__(self, arg): # pragma : no cover
        self._d = dict(arg)

    def __iter__(self):                  # pragma : no cover
        return iter(self._d)

    def __len__(self):                   # pragma : no cover
        return len(self._d)

    def __getitem__(self, key):          # pragma : no cover
        return self._d[key]

    def __repr__(self) :
        return u'<frozendict %s>' % repr(self._d)

    def __hash__(self):
        try:
            return self._cached_hash
        except AttributeError:
            h = self._cached_hash = hash(frozenset(self._d.items()))
            return h


def cartesian(arrays, out=None): # pragma : no cover
    """Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
    1-D arrays to form the cartesian product of.
    out : ndarray
    Array to place the cartesian product in.
    
    Returns
    -------
    out : ndarray
    2-D array of shape (M, len(arrays)) containing cartesian products
    formed of input arrays.
    
    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
    [1, 4, 7],
    [1, 5, 6],
    [1, 5, 7],
    [2, 4, 6],
    [2, 4, 7],
    [2, 5, 6],
    [2, 5, 7],
    [3, 4, 6],
    [3, 4, 7],
    [3, 5, 6],
    [3, 5, 7]])
    
    References
    ----------
    http://stackoverflow.com/q/1208118
    
    """
    arrays = [numpy.asarray(x).ravel() for x in arrays]
    dtype = arrays[0].dtype

    n = numpy.prod([x.size for x in arrays])
    if out is None:
        out = numpy.empty([n, len(arrays)], dtype=dtype)

    m = n / arrays[0].size
    out[:, 0] = numpy.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m, 1:])
        for j in range(1, arrays[0].size):
            out[j * m:(j + 1) * m, 1:] = out[0:m, 1:]
    return out
